# Copyright (c) 2020-2023 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Unit tests for the dependency tree class and dependencies."""

import os
import unittest

from satpy.dependency_tree import DependencyTree
from satpy.tests.utils import make_cid, make_dataid


class TestDependencyTree(unittest.TestCase):
    """Test the dependency tree.

    This is what we are working with::

        None (No Data)
         +DataID(name='comp19')
         + +DataID(name='ds5', resolution=250, modifiers=('res_change',))
         + + +DataID(name='ds5', resolution=250, modifiers=())
         + + +__EMPTY_LEAF_SENTINEL__ (No Data)
         + +DataID(name='comp13')
         + + +DataID(name='ds5', resolution=250, modifiers=('res_change',))
         + + + +DataID(name='ds5', resolution=250, modifiers=())
         + + + +__EMPTY_LEAF_SENTINEL__ (No Data)
         + +DataID(name='ds2', resolution=250, calibration=<calibration.reflectance>, modifiers=())

    """

    def setUp(self):
        """Set up the test tree."""
        self.dependency_tree = DependencyTree(None, None, None)

        composite_1 = make_cid(name="comp19")
        dependency_1 = make_dataid(name="ds5", resolution=250, modifiers=("res_change",))
        dependency_1_1 = make_dataid(name="ds5", resolution=250, modifiers=tuple())
        node_composite_1 = self.dependency_tree.add_leaf(composite_1)
        node_dependency_1 = self.dependency_tree.add_leaf(dependency_1, node_composite_1)
        self.dependency_tree.add_leaf(dependency_1_1, node_dependency_1)
        # ToDo: do we really want then empty node to be at the same level as the unmodified data?
        node_dependency_1.add_child(self.dependency_tree.empty_node)

        dependency_2 = make_cid(name="comp13")
        dependency_2_1 = dependency_1
        node_dependency_2 = self.dependency_tree.add_leaf(dependency_2, node_composite_1)
        self.dependency_tree.add_leaf(dependency_2_1, node_dependency_2)
        # We don't need to add the unmodified dependency a second time.

        dependency_3 = make_dataid(name="ds2", resolution=250, calibration="reflectance", modifiers=tuple())
        self.dependency_tree.add_leaf(dependency_3, node_composite_1)

    @staticmethod
    def _nodes_equal(node_list1, node_list2):
        names1 = [node.name for node in node_list1]
        names2 = [node.name for node in node_list2]
        return sorted(names1) == sorted(names2)

    def test_copy_preserves_all_nodes(self):
        """Test that dependency tree copy preserves all nodes."""
        new_dependency_tree = self.dependency_tree.copy()
        assert self.dependency_tree.empty_node is new_dependency_tree.empty_node
        assert self._nodes_equal(self.dependency_tree.leaves(),
                                 new_dependency_tree.leaves())
        assert self._nodes_equal(self.dependency_tree.trunk(),
                                 new_dependency_tree.trunk())

        # make sure that we can get access to sub-nodes
        c13_id = make_cid(name="comp13")
        assert self._nodes_equal(self.dependency_tree.trunk(limit_nodes_to=[c13_id]),
                                 new_dependency_tree.trunk(limit_nodes_to=[c13_id]))

    def test_copy_preserves_unique_empty_node(self):
        """Test that dependency tree copy preserves the uniqueness of the empty node."""
        new_dependency_tree = self.dependency_tree.copy()
        assert self.dependency_tree.empty_node is new_dependency_tree.empty_node

        assert self.dependency_tree._root.children[0].children[0].children[1] is self.dependency_tree.empty_node
        assert new_dependency_tree._root.children[0].children[0].children[1] is self.dependency_tree.empty_node

    def test_new_dependency_tree_preserves_unique_empty_node(self):
        """Test that dependency tree instantiation preserves the uniqueness of the empty node."""
        new_dependency_tree = DependencyTree(None, None, None)
        assert self.dependency_tree.empty_node is new_dependency_tree.empty_node


class TestMissingDependencies(unittest.TestCase):
    """Test the MissingDependencies exception."""

    def test_new_missing_dependencies(self):
        """Test new MissingDependencies."""
        from satpy.node import MissingDependencies
        error = MissingDependencies("bla")
        assert error.missing_dependencies == "bla"

    def test_new_missing_dependencies_with_message(self):
        """Test new MissingDependencies with a message."""
        from satpy.node import MissingDependencies
        error = MissingDependencies("bla", "This is a message")
        assert "This is a message" in str(error)


class TestMultipleResolutionSameChannelDependency(unittest.TestCase):
    """Test that MODIS situations where the same channel is available at multiple resolution works."""

    def test_modis_overview_1000m(self):
        """Test a modis overview dependency calculation with resolution fixed to 1000m."""
        from satpy import DataQuery
        from satpy._config import PACKAGE_CONFIG_PATH
        from satpy.composites.core import GenericCompositor
        from satpy.dataset import DatasetDict
        from satpy.modifiers.geometry import SunZenithCorrector
        from satpy.readers.core.yaml_reader import FileYAMLReader

        config_file = os.path.join(PACKAGE_CONFIG_PATH, "readers", "modis_l1b.yaml")
        self.reader_instance = FileYAMLReader.from_config_files(config_file)

        overview = {"_satpy_id": make_dataid(name="overview"),
                    "name": "overview",
                    "optional_prerequisites": [],
                    "prerequisites": [DataQuery(name="1", modifiers=("sunz_corrected",)),
                                      DataQuery(name="2", modifiers=("sunz_corrected",)),
                                      DataQuery(name="31")],
                    "standard_name": "overview"}
        compositors = {"modis": DatasetDict()}
        compositors["modis"]["overview"] = GenericCompositor(**overview)

        modifiers = {"modis": {"sunz_corrected": (SunZenithCorrector,
                                                  {"optional_prerequisites": ["solar_zenith_angle"],
                                                   "name": "sunz_corrected",
                                                   "prerequisites": []})}}
        dep_tree = DependencyTree({"modis_l1b": self.reader_instance}, compositors, modifiers)
        dep_tree.populate_with_keys({"overview"}, DataQuery(resolution=1000))
        for key in dep_tree._all_nodes.keys():
            assert key.get("resolution", 1000) == 1000


class TestMultipleSensors(unittest.TestCase):
    """Test cases where multiple sensors are available.

    This is what we are working with::

        None (No Data)
         +DataID(name='comp19')
         + +DataID(name='ds5', resolution=250, modifiers=('res_change',))
         + + +DataID(name='ds5', resolution=250, modifiers=())
         + + +__EMPTY_LEAF_SENTINEL__ (No Data)
         + +DataID(name='comp13')
         + + +DataID(name='ds5', resolution=250, modifiers=('res_change',))
         + + + +DataID(name='ds5', resolution=250, modifiers=())
         + + + +__EMPTY_LEAF_SENTINEL__ (No Data)
         + +DataID(name='ds2', resolution=250, calibration=<calibration.reflectance>, modifiers=())

    """

    def setUp(self):
        """Set up the test tree."""
        from satpy.composites.core import CompositeBase
        from satpy.dataset.data_dict import DatasetDict
        from satpy.modifiers import ModifierBase

        class _FakeCompositor(CompositeBase):
            def __init__(self, ret_val, *args, **kwargs):
                self.ret_val = ret_val
                super().__init__(*args, **kwargs)

            def __call__(self, *args, **kwargs):
                return self.ret_val

        class _FakeModifier(ModifierBase):
            def __init__(self, ret_val, *args, **kwargs):
                self.ret_val = ret_val
                super().__init__(*args, **kwargs)

            def __call__(self, *args, **kwargs):
                return self.ret_val

        comp1_sensor1 = _FakeCompositor(1, "comp1")
        comp1_sensor2 = _FakeCompositor(2, "comp1")
        # create the dictionary one element at a time to force "incorrect" order
        # (sensor2 comes before sensor1, but results should be alphabetical order)
        compositors = {}
        compositors["sensor2"] = s2_comps = DatasetDict()
        compositors["sensor1"] = s1_comps = DatasetDict()
        c1_s2_id = make_cid(name="comp1", resolution=1000)
        c1_s1_id = make_cid(name="comp1", resolution=500)
        s2_comps[c1_s2_id] = comp1_sensor2
        s1_comps[c1_s1_id] = comp1_sensor1

        modifiers = {}
        modifiers["sensor2"] = s2_mods = {}
        modifiers["sensor1"] = s1_mods = {}
        s2_mods["mod1"] = (_FakeModifier, {"ret_val": 2})
        s1_mods["mod1"] = (_FakeModifier, {"ret_val": 1})

        self.dependency_tree = DependencyTree({}, compositors, modifiers)
        # manually add a leaf so we don't have to mock a reader
        ds5 = make_dataid(name="ds5", resolution=250, modifiers=tuple())
        self.dependency_tree.add_leaf(ds5)

    def test_compositor_loaded_sensor_order(self):
        """Test that a compositor is loaded from the first alphabetical sensor."""
        self.dependency_tree.populate_with_keys({"comp1"})
        comp_nodes = self.dependency_tree.trunk()
        assert len(comp_nodes) == 1
        assert comp_nodes[0].name["resolution"] == 500

    def test_modifier_loaded_sensor_order(self):
        """Test that a modifier is loaded from the first alphabetical sensor."""
        from satpy import DataQuery
        dq = DataQuery(name="ds5", modifiers=("mod1",))
        self.dependency_tree.populate_with_keys({dq})
        comp_nodes = self.dependency_tree.trunk()
        assert len(comp_nodes) == 1
        assert comp_nodes[0].data[0].ret_val == 1



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""The ami_l1b reader tests package."""
import contextlib
from typing import Iterator
from unittest import mock

import dask.array as da
import numpy as np
import pytest
import xarray as xr
from pytest import approx, raises  # noqa: PT013

from satpy.readers.ami_l1b import AMIL1bNetCDF
from satpy.tests.utils import make_dataid

FAKE_VIS_DATA = (((np.arange(10.).reshape((2, 5)) + 1.) * 50.0 + 1.0) / 0.5).astype(np.uint16)
FAKE_IR_DATA = ((np.arange(10).reshape((2, 5))) + 7000).astype(np.uint16)


class FakeDataset(object):
    """Mimic xarray Dataset object."""

    def __init__(self, info, attrs):
        """Initialize test data."""
        for var_name, var_data in list(info.items()):
            if isinstance(var_data, np.ndarray):
                info[var_name] = xr.DataArray(var_data)
        self.info = info
        self.attrs = attrs

    def __getitem__(self, key):
        """Mimic getitem method."""
        return self.info[key]

    def __contains__(self, key):
        """Mimic contains method."""
        return key in self.info

    def rename(self, *args, **kwargs):
        """Mimic rename method."""
        return self

    def close(self):
        """Act like close method."""
        return


def _get_fake_counts(rad_data: np.ndarray, attrs: dict) -> xr.DataArray:
    counts = xr.DataArray(
        da.from_array(rad_data, chunks="auto"),
        dims=("y", "x"),
        attrs=attrs,
    )
    return counts


@contextlib.contextmanager
def _fake_reader(counts_data: xr.DataArray) -> Iterator[AMIL1bNetCDF]:
    sc_position = xr.DataArray(0., attrs={
        "sc_position_center_pixel": [-26113466.1974016, 33100139.1630508, 3943.75470244799],
    })
    fake_ds = FakeDataset(
        {
            "image_pixel_values": counts_data,
            "sc_position": sc_position,
            "gsics_coeff_intercept": [0.1859369],
            "gsics_coeff_slope": [0.9967594],
        },
        {
            "satellite_name": "GK-2A",
            "observation_start_time": 623084431.957882,
            "observation_end_time": 623084975.606133,
            "projection_type": "GEOS",
            "sub_longitude": 2.23751210105673,
            "cfac": 81701355.6133574,
            "lfac": -81701355.6133574,
            "coff": 11000.5,
            "loff": 11000.5,
            "nominal_satellite_height": 42164000.,
            "earth_equatorial_radius": 6378137.,
            "earth_polar_radius": 6356752.3,
            "number_of_columns": 22000,
            "number_of_lines": 22000,
            "observation_mode": "FD",
            "channel_spatial_resolution": "0.5",
            "Radiance_to_Albedo_c": 1,
            "DN_to_Radiance_Gain": -0.0144806550815701,
            "DN_to_Radiance_Offset": 118.050903320312,
            "Teff_to_Tbb_c0": -0.141418528203155,
            "Teff_to_Tbb_c1": 1.00052232906885,
            "Teff_to_Tbb_c2": -0.00000036287276076109,
            "light_speed": 2.9979245800E+08,
            "Boltzmann_constant_k": 1.3806488000E-23,
            "Plank_constant_h": 6.6260695700E-34,
        }
    )
    with mock.patch("satpy.readers.ami_l1b.xr") as xr_:
        xr_.open_dataset.return_value = fake_ds
        yield AMIL1bNetCDF("filename",
                            {"platform_shortname": "gk2a"},
                            {"file_type": "ir087"})


@pytest.fixture
def fake_vis_reader():
    """Create fake reader for loading visible data."""
    attrs = _fake_vis_attrs()
    counts_data_arr = _get_fake_counts(FAKE_VIS_DATA, attrs)
    with _fake_reader(counts_data_arr) as reader:
        yield reader


def _fake_vis_attrs():
    return {
        "channel_name": "VI006",
        "detector_side": 2,
        "number_of_total_pixels": 484000000,
        "number_of_error_pixels": 113892451,
        "max_pixel_value": 32768,
        "min_pixel_value": 6,
        "average_pixel_value": 8228.98770845248,
        "stddev_pixel_value": 13621.130386551,
        "number_of_total_bits_per_pixel": 16,
        "number_of_data_quality_flag_bits_per_pixel": 2,
        "number_of_valid_bits_per_pixel": np.array([12]).astype(np.uint8),
        "data_quality_flag_meaning":
            "0:good_pixel, 1:conditionally_usable_pixel, 2:out_of_scan_area_pixel, 3:error_pixel",
        "ground_sample_distance_ew": 1.4e-05,
        "ground_sample_distance_ns": 1.4e-05,
    }


@pytest.fixture
def fake_ir_reader():
    """Create fake reader for loading IR data."""
    attrs = _fake_ir_attrs()
    counts_data_arr = _get_fake_counts(FAKE_IR_DATA, attrs)
    with _fake_reader(counts_data_arr) as reader:
        yield reader


def _fake_ir_attrs():
    return {
        "channel_name": "IR087",
        "detector_side": 2,
        "number_of_total_pixels": 484000000,
        "number_of_error_pixels": 113892451,
        "max_pixel_value": 32768,
        "min_pixel_value": 6,
        "average_pixel_value": 8228.98770845248,
        "stddev_pixel_value": 13621.130386551,
        "number_of_total_bits_per_pixel": 16,
        "number_of_data_quality_flag_bits_per_pixel": 2,
        "number_of_valid_bits_per_pixel": np.array([13]).astype(np.uint8),
        "data_quality_flag_meaning":
            "0:good_pixel, 1:conditionally_usable_pixel, 2:out_of_scan_area_pixel, 3:error_pixel",
        "ground_sample_distance_ew": 1.4e-05,
        "ground_sample_distance_ns": 1.4e-05,
    }


@pytest.fixture
def fake_ir_reader2():
    """Create fake reader for testing radiance clipping."""
    counts_arr = FAKE_IR_DATA.copy()
    counts_arr[0, 0] = 16364
    attrs = _fake_ir_attrs()
    counts_data_arr = _get_fake_counts(counts_arr, attrs)
    with _fake_reader(counts_data_arr) as reader:
        yield reader


class TestAMIL1bNetCDF:
    """Test the AMI L1b reader."""

    def _check_orbital_parameters(self, orb_params):
        """Check that orbital parameters match expected values."""
        exp_params = {
            "projection_altitude": 35785863.0,
            "projection_latitude": 0.0,
            "projection_longitude": 128.2,
            "satellite_actual_altitude": 35782654.56070405,
            "satellite_actual_latitude": 0.005364927,
            "satellite_actual_longitude": 128.2707,
        }
        for key, val in exp_params.items():
            assert val == approx(orb_params[key], abs=1e-3)

    def test_filename_grouping(self):
        """Test that filenames are grouped properly."""
        from satpy.readers.core.grouping import group_files
        filenames = [
            "gk2a_ami_le1b_ir087_fd020ge_201909300300.nc",
            "gk2a_ami_le1b_ir096_fd020ge_201909300300.nc",
            "gk2a_ami_le1b_ir105_fd020ge_201909300300.nc",
            "gk2a_ami_le1b_ir112_fd020ge_201909300300.nc",
            "gk2a_ami_le1b_ir123_fd020ge_201909300300.nc",
            "gk2a_ami_le1b_ir133_fd020ge_201909300300.nc",
            "gk2a_ami_le1b_nr013_fd020ge_201909300300.nc",
            "gk2a_ami_le1b_nr016_fd020ge_201909300300.nc",
            "gk2a_ami_le1b_sw038_fd020ge_201909300300.nc",
            "gk2a_ami_le1b_vi004_fd010ge_201909300300.nc",
            "gk2a_ami_le1b_vi005_fd010ge_201909300300.nc",
            "gk2a_ami_le1b_vi006_fd005ge_201909300300.nc",
            "gk2a_ami_le1b_vi008_fd010ge_201909300300.nc",
            "gk2a_ami_le1b_wv063_fd020ge_201909300300.nc",
            "gk2a_ami_le1b_wv069_fd020ge_201909300300.nc",
            "gk2a_ami_le1b_wv073_fd020ge_201909300300.nc"]
        groups = group_files(filenames, reader="ami_l1b")
        assert len(groups) == 1
        assert len(groups[0]["ami_l1b"]) == 16

    def test_basic_attributes(self, fake_vis_reader):
        """Test getting basic file attributes."""
        import datetime as dt
        assert fake_vis_reader.start_time == dt.datetime(2019, 9, 30, 3, 0, 31, 957882)
        assert fake_vis_reader.end_time == dt.datetime(2019, 9, 30, 3, 9, 35, 606133)

    def test_get_dataset(self, fake_vis_reader):
        """Test getting radiance data."""
        from satpy.tests.utils import make_dataid
        key = make_dataid(name="VI006", calibration="radiance")
        res = fake_vis_reader.get_dataset(key, {
            "file_key": "image_pixel_values",
            "standard_name": "toa_outgoing_radiance_per_unit_wavelength",
            "units": "W m-2 um-1 sr-1",
        })
        exp = {"calibration": "radiance",
               "modifiers": (),
               "platform_name": "GEO-KOMPSAT-2A",
               "sensor": "ami",
               "units": "W m-2 um-1 sr-1"}
        for key, val in exp.items():
            assert val == res.attrs[key]
        self._check_orbital_parameters(res.attrs["orbital_parameters"])

    def test_bad_calibration(self):
        """Test that asking for a bad calibration fails."""
        from satpy.tests.utils import make_dataid
        with raises(ValueError, match="_bad_ invalid value for .*"):
            _ = make_dataid(name="VI006", calibration="_bad_")

    @mock.patch("satpy.readers.core.abi.geometry.AreaDefinition")
    def test_get_area_def(self, adef, fake_vis_reader):
        """Test the area generation."""
        fake_vis_reader.get_area_def(None)

        assert adef.call_count == 1
        call_args = tuple(adef.call_args)[0]
        exp = {"a": 6378137.0, "b": 6356752.3, "h": 35785863.0,
               "lon_0": 128.2, "proj": "geos", "units": "m"}
        for key, val in exp.items():
            assert key in call_args[3]
            assert val == approx(call_args[3][key])
        assert call_args[4] == fake_vis_reader.nc.attrs["number_of_columns"]
        assert call_args[5] == fake_vis_reader.nc.attrs["number_of_lines"]
        np.testing.assert_allclose(call_args[6],
                                   [-5511022.902, -5511022.902, 5511022.902, 5511022.902])

    def test_get_dataset_vis(self, fake_vis_reader):
        """Test get visible calibrated data."""
        from satpy.tests.utils import make_dataid
        key = make_dataid(name="VI006", calibration="reflectance")
        res = fake_vis_reader.get_dataset(key, {
            "file_key": "image_pixel_values",
            "standard_name": "toa_bidirectional_reflectance",
            "units": "%",
        })
        exp = {"calibration": "reflectance",
               "modifiers": (),
               "platform_name": "GEO-KOMPSAT-2A",
               "sensor": "ami",
               "units": "%"}
        for key, val in exp.items():
            assert val == res.attrs[key]
        self._check_orbital_parameters(res.attrs["orbital_parameters"])

    def test_get_dataset_counts(self, fake_vis_reader):
        """Test get counts data."""
        from satpy.tests.utils import make_dataid
        key = make_dataid(name="VI006", calibration="counts")
        res = fake_vis_reader.get_dataset(key, {
            "file_key": "image_pixel_values",
            "standard_name": "counts",
            "units": "1",
        })
        exp = {"calibration": "counts",
               "modifiers": (),
               "platform_name": "GEO-KOMPSAT-2A",
               "sensor": "ami",
               "units": "1"}
        for key, val in exp.items():
            assert val == res.attrs[key]
        self._check_orbital_parameters(res.attrs["orbital_parameters"])


class TestAMIL1bNetCDFIRCal:
    """Test IR specific things about the AMI reader."""
    ds_id = make_dataid(name="IR087", wavelength=[8.415, 8.59, 8.765],
                        calibration="brightness_temperature")
    ds_info = {
        "file_key": "image_pixel_values",
        "wavelength": [8.415, 8.59, 8.765],
        "standard_name": "toa_brightness_temperature",
        "units": "K",
    }

    def test_default_calibrate(self, fake_ir_reader):
        """Test default (pyspectral) IR calibration."""
        from satpy.readers.ami_l1b import rad2temp
        with mock.patch("satpy.readers.ami_l1b.rad2temp", wraps=rad2temp) as r2t_mock:
            res = fake_ir_reader.get_dataset(self.ds_id, self.ds_info)
            r2t_mock.assert_called_once()
        expected = np.array([[238.34385135, 238.31443527, 238.28500087, 238.25554813, 238.22607701],
                             [238.1965875, 238.16707956, 238.13755317, 238.10800829, 238.07844489]])
        np.testing.assert_allclose(res.data.compute(), expected, equal_nan=True)
        # make sure the attributes from the file are in the data array
        assert res.attrs["standard_name"] == "toa_brightness_temperature"

    def test_infile_calibrate(self, fake_ir_reader):
        """Test IR calibration using in-file coefficients."""
        from satpy.readers.ami_l1b import rad2temp
        fake_ir_reader.calib_mode = "FILE"
        with mock.patch("satpy.readers.ami_l1b.rad2temp", wraps=rad2temp) as r2t_mock:
            res = fake_ir_reader.get_dataset(self.ds_id, self.ds_info)
            r2t_mock.assert_not_called()
        expected = np.array([[238.34385135, 238.31443527, 238.28500087, 238.25554813, 238.22607701],
                             [238.1965875, 238.16707956, 238.13755317, 238.10800829, 238.07844489]])
        # file coefficients are pretty close, give some wiggle room
        np.testing.assert_allclose(res.data.compute(), expected, equal_nan=True, atol=0.04)
        # make sure the attributes from the file are in the data array
        assert res.attrs["standard_name"] == "toa_brightness_temperature"

    def test_gsics_radiance_corr(self, fake_ir_reader):
        """Test IR radiance adjustment using in-file GSICS coefs."""
        from satpy.readers.ami_l1b import rad2temp
        fake_ir_reader.calib_mode = "GSICS"
        expected = np.array([[238.036797, 238.007106, 237.977396, 237.947668, 237.91792],
                             [237.888154, 237.85837, 237.828566, 237.798743, 237.768902]])
        with mock.patch("satpy.readers.ami_l1b.rad2temp", wraps=rad2temp) as r2t_mock:
            res = fake_ir_reader.get_dataset(self.ds_id, self.ds_info)
            r2t_mock.assert_not_called()
        # file coefficients are pretty close, give some wiggle room
        np.testing.assert_allclose(res.data.compute(), expected, equal_nan=True, atol=0.01)
        # make sure the attributes from the file are in the data array
        assert res.attrs["standard_name"] == "toa_brightness_temperature"

    def test_user_radiance_corr(self, fake_ir_reader):
        """Test IR radiance adjustment using user-supplied coefs."""
        from satpy.readers.ami_l1b import rad2temp
        fake_ir_reader.calib_mode = "FILE"
        fake_ir_reader.user_calibration = {"IR087": {"slope": 0.99669,
                                                  "offset": 0.16907}}
        expected = np.array([[238.073713, 238.044043, 238.014354, 237.984647, 237.954921],
                             [237.925176, 237.895413, 237.865631, 237.835829, 237.806009]])
        with mock.patch("satpy.readers.ami_l1b.rad2temp", wraps=rad2temp) as r2t_mock:
            res = fake_ir_reader.get_dataset(self.ds_id, self.ds_info)
            r2t_mock.assert_not_called()
        # file coefficients are pretty close, give some wiggle room
        np.testing.assert_allclose(res.data.compute(), expected, equal_nan=True, atol=0.01)
        # make sure the attributes from the file are in the data array
        assert res.attrs["standard_name"] == "toa_brightness_temperature"

    @pytest.mark.parametrize("clip", [False, True])
    def test_clipneg(self, fake_ir_reader2, clip):
        """Test that negative radiances are clipped."""
        ds_id = make_dataid(name="IR087", wavelength=[8.415, 8.59, 8.765],
                            calibration="radiance")
        fake_ir_reader2.clip_negative_radiances = clip
        res = np.array(fake_ir_reader2.get_dataset(ds_id, self.ds_info))
        if clip:
            np.testing.assert_allclose(res[0, 0], 0.004603, atol=0.0001)
        else:
            assert res[0, 0] < 0



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2010-2017 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""HRIT format reader for JMA data.

The JMA HRIT format is described in the `JMA HRIT - Mission Specific
Implementation`_.


Example:
--------
Here is an example how to read Himwari-8 HRIT data with Satpy:

.. code-block:: python

    from satpy import Scene
    import glob

    filenames = glob.glob('data/IMG_DK01B14_2018011109*')
    scn = Scene(filenames=filenames, reader='ahi_hrit')
    scn.load(['B14'])
    print(scn['B14'])


Output:

.. code-block:: none

    <xarray.DataArray (y: 5500, x: 5500)>
    dask.array<concatenate, shape=(5500, 5500), dtype=float64, chunksize=(550, 4096), ...
    Coordinates:
        acq_time  (y) datetime64[ns] 2018-01-11T09:00:20.995200 ... 2018-01-11T09:09:40.348800
        crs       object +proj=geos +lon_0=140.7 +h=35785831 +x_0=0 +y_0=0 +a=6378169 ...
      * y         (y) float64 5.5e+06 5.498e+06 5.496e+06 ... -5.496e+06 -5.498e+06
      * x         (x) float64 -5.498e+06 -5.496e+06 -5.494e+06 ... 5.498e+06 5.5e+06
    Attributes:
        orbital_parameters:   {'projection_longitude': 140.7, 'projection_latitud...
        standard_name:        toa_brightness_temperature
        level:                None
        wavelength:           (11.0, 11.2, 11.4)
        units:                K
        calibration:          brightness_temperature
        file_type:            ['hrit_b14_seg', 'hrit_b14_fd']
        modifiers:            ()
        polarization:         None
        sensor:               ahi
        name:                 B14
        platform_name:        Himawari-8
        resolution:           4000
        start_time:           2018-01-11 09:00:20.995200
        end_time:             2018-01-11 09:09:40.348800
        area:                 Area ID: FLDK, Description: Full Disk, Projection I...
        ancillary_variables:  []

JMA HRIT data contain the scanline acquisition time for only a subset of scanlines. Timestamps of
the remaining scanlines are computed using linear interpolation. This is what you'll find in the
``acq_time`` coordinate of the dataset.

Compression
-----------

Gzip-compressed MTSAT files can be decompressed on the fly using
:class:`~satpy.readers.core.remote.FSFile`:

.. code-block:: python

    import fsspec
    from satpy import Scene
    from satpy.readers.core.remote import FSFile

    filename = "/data/HRIT_MTSAT1_20090101_0630_DK01IR1.gz"
    open_file = fsspec.open(filename, compression="gzip")
    fs_file = FSFile(open_file)
    scn = Scene([fs_file], reader="jami_hrit")
    scn.load(["IR1"])


.. _JMA HRIT - Mission Specific Implementation: http://www.jma.go.jp/jma/jma-eng/satellite/introduction/4_2HRIT.pdf
"""

import datetime as dt
import logging

import numpy as np
import xarray as xr

import satpy.utils
from satpy.readers.core._geos_area import get_area_definition, get_area_extent
from satpy.readers.core.hrit import (
    HRITFileHandler,
    ancillary_text,
    annotation_header,
    base_hdr_map,
    image_data_function,
)
from satpy.readers.core.utils import get_geostationary_mask

logger = logging.getLogger("hrit_jma")


# JMA implementation:
key_header = np.dtype([("key_number", "u4")])

segment_identification = np.dtype([("image_segm_seq_no", ">u1"),
                                   ("total_no_image_segm", ">u1"),
                                   ("line_no_image_segm", ">u2")])

encryption_key_message = np.dtype([("station_number", ">u2")])

image_compensation_information = np.dtype([("compensation", "|S1")])

image_observation_time = np.dtype([("times", "|S1")])

image_quality_information = np.dtype([("quality", "|S1")])


jma_variable_length_headers: dict = {}

jma_text_headers = {image_data_function: "image_data_function",
                    annotation_header: "annotation_header",
                    ancillary_text: "ancillary_text",
                    image_compensation_information: "image_compensation_information",
                    image_observation_time: "image_observation_time",
                    image_quality_information: "image_quality_information"}

jma_hdr_map = base_hdr_map.copy()
jma_hdr_map.update({7: key_header,
                    128: segment_identification,
                    129: encryption_key_message,
                    130: image_compensation_information,
                    131: image_observation_time,
                    132: image_quality_information
                    })


cuc_time = np.dtype([("coarse", "u1", (4, )),
                     ("fine", "u1", (3, ))])

time_cds_expanded = np.dtype([("days", ">u2"),
                              ("milliseconds", ">u4"),
                              ("microseconds", ">u2"),
                              ("nanoseconds", ">u2")])

FULL_DISK = 1
NORTH_HEMIS = 2
SOUTH_HEMIS = 3
UNKNOWN_AREA = -1
AREA_NAMES = {FULL_DISK: {"short": "FLDK", "long": "Full Disk"},
              NORTH_HEMIS: {"short": "NH", "long": "Northern Hemisphere"},
              SOUTH_HEMIS: {"short": "SH", "long": "Southern Hemisphere"},
              UNKNOWN_AREA: {"short": "UNKNOWN", "long": "Unknown Area"}}

MTSAT1R = "MTSAT-1R"
MTSAT2 = "MTSAT-2"
HIMAWARI8 = "Himawari-8"
UNKNOWN_PLATFORM = "Unknown Platform"
PLATFORMS = {
    "GEOS(140.00)": MTSAT1R,
    "GEOS(140.25)": MTSAT1R,
    "GEOS(140.70)": HIMAWARI8,
    "GEOS(145.00)": MTSAT2,
}
SENSORS = {
    MTSAT1R: "jami",
    MTSAT2: "mtsat2_imager",
    HIMAWARI8: "ahi"
}


def mjd2datetime64(mjd):
    """Convert Modified Julian Day (MJD) to datetime64."""
    epoch = np.datetime64("1858-11-17 00:00")
    day2nsec = 24 * 3600 * 1E9
    mjd_nsec = (mjd * day2nsec).astype(np.int64).astype("timedelta64[ns]")

    return epoch + mjd_nsec


class HRITJMAFileHandler(HRITFileHandler):
    """JMA HRIT format reader.

    By default, the reader uses the start time parsed from the filename. To use exact time, computed
    from the metadata, the user can define a keyword argument::

        scene = Scene(filenames=filenames,
                      reader='ahi_hrit',
                      reader_kwargs={'use_acquisition_time_as_start_time': True})

    As this time is different for every channel, time-dependent calculations like SZA correction
    can be pretty slow when multiple channels are used.

    The exact scanline times are always available as coordinates of an individual channels::

        scene.load(["B03"])
        print(scene["B03].coords["acq_time"].data)

    would print something similar to::

        array(['2021-12-08T06:00:20.131200000', '2021-12-08T06:00:20.191948000',
               '2021-12-08T06:00:20.252695000', ...,
               '2021-12-08T06:09:39.449390000', '2021-12-08T06:09:39.510295000',
               '2021-12-08T06:09:39.571200000'], dtype='datetime64[ns]')

    The first value represents the exact start time, and the last one the exact end time of the data
    acquisition.

    """

    def __init__(self, filename, filename_info, filetype_info, use_acquisition_time_as_start_time=False):
        """Initialize the reader."""
        super(HRITJMAFileHandler, self).__init__(filename, filename_info,
                                                 filetype_info,
                                                 (jma_hdr_map,
                                                  jma_variable_length_headers,
                                                  jma_text_headers))

        self._use_acquisition_time_as_start_time = use_acquisition_time_as_start_time
        self.mda["segment_sequence_number"] = self.mda["image_segm_seq_no"]
        self.mda["planned_end_segment_number"] = self.mda["total_no_image_segm"]
        self.mda["planned_start_segment_number"] = np.uint8(1)

        items = self.mda["image_data_function"].decode().split("\r")
        if items[0].startswith("$HALFTONE"):
            self.calibration_table = []
            for item in items[1:]:
                if item == "":
                    continue
                key, value = item.split(":=")
                if key.startswith("_UNIT"):
                    self.mda["unit"] = item.split(":=")[1]
                elif key.startswith("_NAME"):
                    pass
                elif key.isdigit():
                    key = int(key)
                    value = float(value)
                    self.calibration_table.append((key, value))

            self.calibration_table = np.array(self.calibration_table)

        self.projection_name = self.mda["projection_name"].decode().strip()
        sublon = float(self.projection_name.split("(")[1][:-1])
        self.mda["projection_parameters"]["SSP_longitude"] = sublon
        self.platform = self._get_platform()
        self.is_segmented = self.mda["segment_sequence_number"] > 0
        self.area_id = filename_info.get("area", UNKNOWN_AREA)
        if self.area_id not in AREA_NAMES:
            self.area_id = UNKNOWN_AREA
        self.area = self._get_area_def()
        self.acq_time = self._get_acq_time()

    def _get_platform(self):
        """Get the platform name.

        The platform is not specified explicitly in JMA HRIT files. For
        segmented data it is not even specified in the filename. But it
        can be derived indirectly from the projection name:

            GEOS(140.00): MTSAT-1R
            GEOS(140.25): MTSAT-1R    # TODO: Check if there is more...
            GEOS(140.70): Himawari-8
            GEOS(145.00): MTSAT-2

        See [MTSAT], section 3.1. Unfortunately Himawari-8 and 9 are not
        distinguishable using that method at the moment. From [HIMAWARI]:

        "HRIT/LRIT files have the same file naming convention in the same
        format in Himawari-8 and Himawari-9, so there is no particular
        difference."

        TODO: Find another way to distinguish Himawari-8 and 9.

        References:
        [MTSAT] http://www.data.jma.go.jp/mscweb/notice/Himawari7_e.html
        [HIMAWARI] http://www.data.jma.go.jp/mscweb/en/himawari89/space_segment/sample_hrit.html

        """
        try:
            return PLATFORMS[self.projection_name]
        except KeyError:
            logger.error("Unable to determine platform: Unknown projection "
                         'name "{}"'.format(self.projection_name))
            return UNKNOWN_PLATFORM

    def _check_sensor_platform_consistency(self, sensor):
        """Make sure sensor and platform are consistent.

        Args:
            sensor (str) : Sensor name from YAML dataset definition

        Raises:
            (Exception) ValueError: if they don't match

        """
        ref_sensor = SENSORS.get(self.platform, None)
        if ref_sensor and not sensor == ref_sensor:
            logger.error("Sensor-Platform mismatch: {} is not a payload "
                         "of {}. Did you choose the correct reader?"
                         .format(sensor, self.platform))

    def _get_line_offset(self):
        """Get line offset for the current segment.

        Read line offset from the file and adapt it to the current segment
        or half disk scan so that

            y(l) ~ l - loff

        because this is what get_geostationary_area_extent() expects.
        """
        # Get line offset from the file
        nlines = int(self.mda["number_of_lines"])
        loff = np.float32(self.mda["loff"])

        # Adapt it to the current segment
        if self.is_segmented:
            # loff in the file specifies the offset of the full disk image
            # centre (1375/2750 for VIS/IR)
            segment_number = int(self.mda["segment_sequence_number"]) - 1
            loff -= (int(self.mda["total_no_image_segm"]) - segment_number - 1) * nlines
        elif self.area_id in (NORTH_HEMIS, SOUTH_HEMIS):
            # loff in the file specifies the start line of the half disk image
            # in the full disk image
            loff = nlines - loff
        elif self.area_id == UNKNOWN_AREA:
            logger.error("Cannot compute line offset for unknown area")

        return loff

    def _get_area_def(self):
        """Get the area definition of the band."""
        pdict = {
            "cfac": np.int32(self.mda["cfac"]),
            "lfac": np.int32(self.mda["lfac"]),
            "coff": np.float32(self.mda["coff"]),
            "loff": self._get_line_offset(),
            "ncols": int(self.mda["number_of_columns"]),
            "nlines": int(self.mda["number_of_lines"]),
            "scandir": "N2S",
            "a": float(self.mda["projection_parameters"]["a"]),
            "b": float(self.mda["projection_parameters"]["b"]),
            "h": float(self.mda["projection_parameters"]["h"]),
            "ssp_lon": float(self.mda["projection_parameters"]["SSP_longitude"]),
            "a_name": AREA_NAMES[self.area_id]["short"],
            "a_desc": AREA_NAMES[self.area_id]["long"],
            "p_id": "geosmsg"
        }
        area_extent = get_area_extent(pdict)
        return get_area_definition(pdict, area_extent)

    def get_area_def(self, dsid):
        """Get the area definition of the band."""
        return self.area

    def get_dataset(self, key, info):
        """Get the dataset designated by *key*."""
        res = super(HRITJMAFileHandler, self).get_dataset(key, info)

        # Filenames of segmented data is identical for MTSAT-1R, MTSAT-2
        # and Himawari-8/9. Make sure we have the correct reader for the data
        # at hand.
        self._check_sensor_platform_consistency(info["sensor"])

        # Calibrate and mask space pixels
        res = self._mask_space(self.calibrate(res, key["calibration"]))

        # Add scanline acquisition time
        res.coords["acq_time"] = ("y", self.acq_time)
        res.coords["acq_time"].attrs["long_name"] = "Scanline acquisition time"

        # Update attributes
        res.attrs.update(info)
        res.attrs["platform_name"] = self.platform
        res.attrs["orbital_parameters"] = {
            "projection_longitude": float(self.mda["projection_parameters"]["SSP_longitude"]),
            "projection_latitude": 0.,
            "projection_altitude": float(self.mda["projection_parameters"]["h"])}

        return res

    def _mask_space(self, data):
        """Mask space pixels."""
        geomask = get_geostationary_mask(area=self.area)
        return data.where(geomask)

    def _get_acq_time(self):
        r"""Get the acquisition times from the file.

        Acquisition times for a subset of scanlines are stored in the header
        as follows:

        b'LINE:=1\rTIME:=54365.022558\rLINE:=21\rTIME:=54365.022664\r...'

        Missing timestamps in between are computed using linear interpolation.
        """
        buf_b = np.frombuffer(self.mda["image_observation_time"],
                              dtype=image_observation_time)

        # Replace \r by \n before encoding, otherwise encoding will drop all
        # elements except the last one
        buf_s = b"".join(buf_b["times"]).replace(b"\r", b"\n").decode()

        # Split into key:=value pairs; then extract line number and timestamp
        splits = buf_s.strip().split("\n")
        lines_sparse = [int(s.split(":=")[1]) for s in splits[0::2]]
        times_sparse = [float(s.split(":=")[1]) for s in splits[1::2]]

        if self.platform == HIMAWARI8:
            # Only a couple of timestamps in the header, and only the first
            # and last are usable (duplicates inbetween).
            lines_sparse = [lines_sparse[0], lines_sparse[-1]]
            times_sparse = [times_sparse[0], times_sparse[-1]]

        # Compute missing timestamps using linear interpolation.
        lines = np.arange(lines_sparse[0], lines_sparse[-1]+1)
        times = np.interp(lines, lines_sparse, times_sparse)

        # Convert to np.datetime64
        times64 = mjd2datetime64(times)

        return times64

    @staticmethod
    def _interp(arr, cal):
        return np.interp(arr.ravel(), cal[:, 0], cal[:, 1]).reshape(arr.shape)

    def calibrate(self, data, calibration):
        """Calibrate the data."""
        tic = dt.datetime.now()

        if calibration == "counts":
            return data
        if calibration == "radiance":
            raise NotImplementedError("Can't calibrate to radiance.")

        cal = self.calibration_table
        res = data.data.map_blocks(self._interp, cal, dtype=cal[:, 0].dtype)
        res = xr.DataArray(res,
                           dims=data.dims, attrs=data.attrs,
                           coords=data.coords)
        res = res.where(data < 65535)
        logger.debug("Calibration time " + str(dt.datetime.now() - tic))
        return res

    @property
    def start_time(self):
        """Get start time of the scan."""
        if self._use_acquisition_time_as_start_time:
            return satpy.utils.datetime64_to_pydatetime(self.acq_time[0])
        return self._start_time

    @property
    def end_time(self):
        """Get end time of the scan."""
        return satpy.utils.datetime64_to_pydatetime(self.acq_time[-1])



# Copyright (c) 2022 Pytroll Developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Reader for the EPS-SG Microwave Sounder (MWS) level-1b data.

Documentation: https://www.eumetsat.int/media/44139
"""

import datetime as dt
import logging

import dask.array as da
import numpy as np
from netCDF4 import default_fillvals

from satpy.readers.core.netcdf import NetCDF4FileHandler

logger = logging.getLogger(__name__)


# dict containing all available auxiliary data parameters to be read using the index map. Keys are the
# parameter name and values are the paths to the variable inside the netcdf

AUX_DATA = {
    "scantime_utc": "data/navigation/mws_scantime_utc",
    "solar_azimuth": "data/navigation/mws_solar_azimuth_angle",
    "solar_zenith": "data/navigation/mws_solar_zenith_angle",
    "satellite_azimuth": "data/navigation/mws_satellite_azimuth_angle",
    "satellite_zenith": "data/navigation/mws_satellite_zenith_angle",
    "surface_type": "data/navigation/mws_surface_type",
    "terrain_elevation": "data/navigation/mws_terrain_elevation",
    "mws_lat": "data/navigation/mws_lat",
    "mws_lon": "data/navigation/mws_lon",
}

MWS_CHANNEL_NAMES_TO_NUMBER = {"1": 1, "2": 2, "3": 3, "4": 4,
                               "5": 5, "6": 6, "7": 7, "8": 8,
                               "9": 9, "10": 10, "11": 11, "12": 12,
                               "13": 13, "14": 14, "15": 15, "16": 16,
                               "17": 17, "18": 18, "19": 19, "20": 20,
                               "21": 21, "22": 22, "23": 23, "24": 24}

MWS_CHANNEL_NAMES = list(MWS_CHANNEL_NAMES_TO_NUMBER.keys())
MWS_CHANNELS = set(MWS_CHANNEL_NAMES)


def get_channel_index_from_name(chname):
    """Get the MWS channel index from the channel name."""
    chindex = MWS_CHANNEL_NAMES_TO_NUMBER.get(chname, 0) - 1
    if 0 <= chindex < 24:
        return chindex
    raise AttributeError(f"Channel name {chname!r} not supported")


def _get_aux_data_name_from_dsname(dsname):
    aux_data_name = [key for key in AUX_DATA.keys() if key in dsname]
    if len(aux_data_name) > 0:
        return aux_data_name[0]


class MWSL1BFile(NetCDF4FileHandler):
    """Class implementing the EPS-SG-A1 MWS L1b Filehandler.

    This class implements the European Polar System Second Generation (EPS-SG)
    Microwave Sounder (MWS) Level-1b NetCDF reader.  It is designed to be used
    through the :class:`Scene <satpy.scene.Scene>` class using the :mod:`Scene.load <satpy.scene.Scene.load>`
    method with the reader ``"mws_l1b_nc"``.

    """

    _platform_name_translate = {
        "SGA1": "Metop-SG-A1",
        "SGA2": "Metop-SG-A2",
        "SGA3": "Metop-SG-A3"}

    def __init__(self, filename, filename_info, filetype_info):
        """Initialize file handler."""
        super().__init__(filename, filename_info,
                         filetype_info,
                         cache_var_size=10000,
                         cache_handle=True)
        logger.debug("Reading: {}".format(self.filename))
        logger.debug("Start: {}".format(self.start_time))
        logger.debug("End: {}".format(self.end_time))

        self._cache = {}

        self._channel_names = MWS_CHANNEL_NAMES

    @property
    def start_time(self):
        """Get start time."""
        return dt.datetime.strptime(self["/attr/sensing_start_time_utc"],
                                 "%Y-%m-%d %H:%M:%S.%f")

    @property
    def end_time(self):
        """Get end time."""
        return dt.datetime.strptime(self["/attr/sensing_end_time_utc"],
                                 "%Y-%m-%d %H:%M:%S.%f")

    @property
    def sensor(self):
        """Get the sensor name."""
        return self["/attr/instrument"]

    @property
    def platform_name(self):
        """Get the platform name."""
        return self._platform_name_translate.get(self["/attr/spacecraft"])

    @property
    def sub_satellite_longitude_start(self):
        """Get the longitude of sub-satellite point at start of the product."""
        return self["status/satellite/subsat_longitude_start"].data.item()

    @property
    def sub_satellite_latitude_start(self):
        """Get the latitude of sub-satellite point at start of the product."""
        return self["status/satellite/subsat_latitude_start"].data.item()

    @property
    def sub_satellite_longitude_end(self):
        """Get the longitude of sub-satellite point at end of the product."""
        return self["status/satellite/subsat_longitude_end"].data.item()

    @property
    def sub_satellite_latitude_end(self):
        """Get the latitude of sub-satellite point at end of the product."""
        return self["status/satellite/subsat_latitude_end"].data.item()

    def get_dataset(self, dataset_id, dataset_info):
        """Get dataset using file_key in dataset_info."""
        logger.debug("Reading {} from {}".format(dataset_id["name"], self.filename))

        var_key = dataset_info["file_key"]
        if _get_aux_data_name_from_dsname(dataset_id["name"]) is not None:
            variable = self._get_dataset_aux_data(dataset_id["name"])
        elif any(lb in dataset_id["name"] for lb in MWS_CHANNELS):
            logger.debug(f"Reading in file to get dataset with key {var_key}.")
            variable = self._get_dataset_channel(dataset_id, dataset_info)
        else:
            logger.warning(f"Could not find key {var_key} in NetCDF file, no valid Dataset created")  # noqa: E501
            return None

        variable = self._manage_attributes(variable, dataset_info)
        variable = self._drop_coords(variable)
        variable = self._standardize_dims(variable)
        return variable

    @staticmethod
    def _standardize_dims(variable):
        """Standardize dims to y, x."""
        if "n_scans" in variable.dims:
            variable = variable.rename({"n_fovs": "x", "n_scans": "y"})
        if variable.dims[0] == "x":
            variable = variable.transpose("y", "x")
        return variable

    @staticmethod
    def _drop_coords(variable):
        """Drop coords that are not in dims."""
        for coord in variable.coords:
            if coord not in variable.dims:
                variable = variable.drop_vars(coord)
        return variable

    def _manage_attributes(self, variable, dataset_info):
        """Manage attributes of the dataset."""
        variable.attrs.setdefault("units", None)
        variable.attrs.update(dataset_info)
        variable.attrs.update(self._get_global_attributes())
        return variable

    def _get_dataset_channel(self, key, dataset_info):
        """Load dataset corresponding to channel measurement.

        Load a dataset when the key refers to a measurand, whether uncalibrated
        (counts) or calibrated in terms of brightness temperature or radiance.

        """
        # Get the dataset
        # Get metadata for given dataset
        grp_pth = dataset_info["file_key"]
        channel_index = get_channel_index_from_name(key["name"])

        data = self[grp_pth][:, :, channel_index]
        attrs = data.attrs.copy()

        fv = attrs.pop(
            "FillValue",
            default_fillvals.get(data.dtype.str[1:], np.nan))
        vr = attrs.get("valid_range", [-np.inf, np.inf])

        if key["calibration"] == "counts":
            attrs["_FillValue"] = fv
            nfv = fv
        else:
            nfv = np.nan
        data = data.where(data >= vr[0], nfv)
        data = data.where(data <= vr[1], nfv)

        # Manage the attributes of the dataset
        data.attrs.setdefault("units", None)
        data.attrs.update(dataset_info)

        dataset_attrs = getattr(data, "attrs", {})
        dataset_attrs.update(dataset_info)
        dataset_attrs.update({
            "platform_name": self.platform_name,
            "sensor": self.sensor,
            "orbital_parameters": {"sub_satellite_latitude_start": self.sub_satellite_latitude_start,
                                   "sub_satellite_longitude_start": self.sub_satellite_longitude_start,
                                   "sub_satellite_latitude_end": self.sub_satellite_latitude_end,
                                   "sub_satellite_longitude_end": self.sub_satellite_longitude_end},
        })

        try:
            dataset_attrs.update(key.to_dict())
        except AttributeError:
            dataset_attrs.update(key)

        data.attrs.update(dataset_attrs)
        return data

    def _get_dataset_aux_data(self, dsname):
        """Get the auxiliary data arrays using the index map."""
        # Geolocation and navigation data:
        if dsname in ["mws_lat", "mws_lon",
                      "solar_azimuth", "solar_zenith",
                      "satellite_azimuth", "satellite_zenith",
                      "surface_type", "terrain_elevation"]:
            var_key = AUX_DATA.get(dsname)
        else:
            raise NotImplementedError(f"Dataset {dsname!r} not supported!")

        try:
            variable = self[var_key]
        except KeyError:
            logger.exception("Could not find key %s in NetCDF file, no valid Dataset created", var_key)
            raise

        # Scale the data:
        if "scale_factor" in variable.attrs and "add_offset" in variable.attrs:
            missing_value = variable.attrs["missing_value"]
            variable.data = da.where(variable.data == missing_value, np.nan,
                                     variable.data * variable.attrs["scale_factor"] + variable.attrs["add_offset"])

        return variable

    def _get_global_attributes(self):
        """Create a dictionary of global attributes."""
        return {
            "filename": self.filename,
            "start_time": self.start_time,
            "end_time": self.end_time,
            "spacecraft_name": self.platform_name,
            "sensor": self.sensor,
            "filename_start_time": self.filename_info["start_time"],
            "filename_end_time": self.filename_info["end_time"],
            "platform_name": self.platform_name,
            "quality_group": self._get_quality_attributes(),
        }

    def _get_quality_attributes(self):
        """Get quality attributes."""
        quality_group = self["quality"]
        quality_dict = {}
        for key in quality_group:
            # Add the values (as Numpy array) of each variable in the group
            # where possible
            try:
                quality_dict[key] = quality_group[key].values
            except ValueError:
                quality_dict[key] = None

        quality_dict.update(quality_group.attrs)
        return quality_dict



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2018 - 2020 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
"""Tests for compositors."""



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Tests for the 'cmsaf-claas2_l2_nc' reader."""

import datetime  # noqa: I001
import os

import netCDF4
import numpy as np
import pytest
import xarray as xr
from pyresample.geometry import AreaDefinition

from satpy.tests.utils import make_dataid

# NOTE:
# The following fixtures are not defined in this file, but are used and injected by Pytest:
# - tmp_path
# - request


@pytest.fixture(
    params=[datetime.datetime(2017, 12, 5), datetime.datetime(2017, 12, 6)]
)
def start_time(request):
    """Get start time of the dataset."""
    return request.param


@pytest.fixture
def start_time_str(start_time):
    """Get string representation of the start time."""
    return start_time.strftime("%Y-%m-%dT%H:%M:%SZ")


@pytest.fixture
def fake_dataset(start_time_str):
    """Create a CLAAS-like test dataset."""
    cph = xr.DataArray(
        [[[0, 1], [2, 0]]],
        dims=("time", "y", "x")
    )
    ctt = xr.DataArray(
        [[280, 290], [300, 310]],
        dims=("y", "x")
    )
    time_bounds = xr.DataArray(
        [[12436.91666667, 12436.92534722]],
        dims=("time", "bndsize")
    )
    attrs = {
        "CMSAF_proj4_params": "+a=6378169.0 +h=35785831.0 "
                              "+b=6356583.8 +lon_0=0 +proj=geos",
        "CMSAF_area_extent": np.array(
            [-5456233.41938636, -5453233.01608472,
             5453233.01608472, 5456233.41938636]),
        "time_coverage_start": start_time_str,
        "time_coverage_end": "2085-08-13T13:15:00Z",
    }
    return xr.Dataset(
        {
            "cph": cph,
            "ctt": ctt,
            "time_bnds": time_bounds
        },
        attrs=attrs
    )


@pytest.fixture
def encoding():
    """Dataset encoding."""
    return {
        "ctt": {"scale_factor": np.float32(0.01)},
    }


@pytest.fixture
def fake_file(fake_dataset, encoding, tmp_path):
    """Write a fake dataset to file."""
    filename = tmp_path / "CPPin20140101001500305SVMSG01MD.nc"
    fake_dataset.to_netcdf(filename, encoding=encoding)
    return filename


@pytest.fixture
def fake_files(fake_dataset, encoding, tmp_path):
    """Write the same fake dataset into two different files."""
    filenames = [
        tmp_path / "CPPin20140101001500305SVMSG01MD.nc",
        tmp_path / "CPPin20140101003000305SVMSG01MD.nc",
    ]
    for filename in filenames:
        fake_dataset.to_netcdf(filename, encoding=encoding)
    return filenames


@pytest.fixture
def reader():
    """Return reader for CMSAF CLAAS-2."""
    from satpy._config import config_search_paths
    from satpy.readers.core.loading import load_reader

    reader_configs = config_search_paths(
        os.path.join("readers", "cmsaf-claas2_l2_nc.yaml"))
    reader = load_reader(reader_configs)
    return reader


def test_file_pattern(reader):
    """Test file pattern matching."""
    filenames = [
            "CTXin20040120091500305SVMSG01MD.nc",
            "CTXin20040120093000305SVMSG01MD.nc",
            "CTXin20040120094500305SVMSG01MD.nc",
            "abcde52034294023489248MVSSG03DD.nc"]

    files = reader.select_files_from_pathnames(filenames)
    # only 3 out of 4 above should match
    assert len(files) == 3


class TestCLAAS2MultiFile:
    """Test reading multiple CLAAS-2 files."""

    @pytest.fixture
    def multi_file_reader(self, reader, fake_files):
        """Create a multi-file reader."""
        loadables = reader.select_files_from_pathnames(fake_files)
        reader.create_filehandlers(loadables)
        return reader

    @pytest.fixture
    def multi_file_dataset(self, multi_file_reader):
        """Load datasets from multiple files."""
        ds_ids = [make_dataid(name=name) for name in ["cph", "ctt"]]
        datasets = multi_file_reader.load(ds_ids)
        return datasets

    def test_combine_timestamps(self, multi_file_reader, start_time):
        """Test combination of timestamps."""
        assert multi_file_reader.start_time == start_time
        assert multi_file_reader.end_time == datetime.datetime(2085, 8, 13, 13, 15)

    @pytest.mark.parametrize(
        ("ds_name", "expected"),
        [
            ("cph", [[0, 1], [2, 0], [0, 1], [2, 0]]),
            ("ctt", [[280, 290], [300, 310], [280, 290], [300, 310]]),
        ]
    )
    def test_combine_datasets(self, multi_file_dataset, ds_name, expected):
        """Test combination of datasets."""
        np.testing.assert_array_almost_equal(
            multi_file_dataset[ds_name].data, expected
        )

    def test_number_of_datasets(self, multi_file_dataset):
        """Test number of datasets."""
        assert 2 == len(multi_file_dataset)


class TestCLAAS2SingleFile:
    """Test reading a single CLAAS2 file."""

    @pytest.fixture
    def file_handler(self, fake_file):
        """Return a CLAAS-2 file handler."""
        from satpy.readers.cmsaf_claas2 import CLAAS2
        return CLAAS2(fake_file, {}, {})

    @pytest.mark.parametrize(
        ("ds_name", "expected"),
        [
            ("ctt", xr.DataArray([[280, 290], [300, 310]], dims=("y", "x"))),
            ("cph", xr.DataArray([[0, 1], [2, 0]], dims=("y", "x"))),
        ]
    )
    def test_get_dataset(self, file_handler, ds_name, expected):
        """Test dataset loading."""
        dsid = make_dataid(name=ds_name)
        ds = file_handler.get_dataset(dsid, {})
        xr.testing.assert_allclose(ds, expected)

    def test_start_time(self, file_handler, start_time):
        """Test start time property."""
        assert file_handler.start_time == start_time

    def test_end_time(self, file_handler):
        """Test end time property."""
        assert file_handler.end_time == datetime.datetime(2085, 8, 13, 13, 15)

class TestAreaDef:
    """Test the area of both CLAAS2 and CLAAS3 products."""
    @pytest.fixture(params=[3636, 3712])
    def grid_size(self, request):
        """Return the grid size."""
        return request.param

    @pytest.fixture
    def fake_file(self, tmp_path, start_time_str, grid_size):
        """Write a fake dataset to file.

        Only add y dimension, so that the file is small but we can still check the
        real-life area definition.
        """
        filename = tmp_path / "CPPin20140101001500305SVMSG01MD.nc"
        ds = netCDF4.Dataset(filename, mode="w")
        ds.createDimension("y", grid_size)
        ds.setncattr("time_coverage_start", start_time_str)
        ds.close()
        return filename

    @pytest.fixture
    def file_handler(self, fake_file):
        """Return a CLAAS-2 file handler."""
        from satpy.readers.cmsaf_claas2 import CLAAS2
        return CLAAS2(fake_file, {}, {})

    @pytest.fixture
    def area_extent_exp(self, start_time, grid_size):
        """Get expected area extent."""
        if start_time < datetime.datetime(2017, 12, 6):
            extents = {
                3636: (-5454733.160460291, -5454733.160460292, 5454733.160460292, 5454733.160460291),
                3712: (-5568748.485046371, -5568748.485046371, 5568748.485046371, 5568748.485046371)
            }
        else:
            extents = {
                3636: (-5456233.362099582, -5453232.958821001, 5453232.958821001, 5456233.362099582),
                3712: (-5570248.686685662, -5567248.28340708, 5567248.28340708, 5570248.686685662)
            }
        return extents[grid_size]

    @pytest.fixture
    def area_exp(self, area_extent_exp, grid_size):
        """Get expected area definition."""
        proj_dict = {
            "a": 6378169.0,
            "b": 6356583.8,
            "h": 35785831.0,
            "lon_0": 0.0,
            "proj": "geos",
            "units": "m",
        }
        return AreaDefinition(
            area_id="msg_seviri_fes_3km",
            description="MSG SEVIRI Full Earth Scanning service area definition with 3 km resolution",
            proj_id="geos",
            projection=proj_dict,
            area_extent=area_extent_exp,
            width=grid_size,
            height=grid_size,
        )

    def test_get_area_def(self, file_handler, area_exp):
        """Test area definition."""
        area = file_handler.get_area_def(make_dataid(name="foo"))
        assert area == area_exp



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2021 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Benchmark ABI L1B operations."""
from __future__ import annotations

import os

from pyspectral.rayleigh import check_and_download as download_luts
from pyspectral.rsr_reader import check_and_download as download_rsr

from benchmarks.utils import GeoBenchmarks, get_filenames


class ABIL1B(GeoBenchmarks):
    """Benchmark ABI L1B reading."""

    timeout = 600
    data_files: list[str] = []
    subdir = os.path.join("abi_l1b", "20190314_us_midlatitude_cyclone")
    reader = "abi_l1b"

    def setup_cache(self):
        """Fetch the data files."""
        try:
            from satpy.demo import get_us_midlatitude_cyclone_abi
            get_us_midlatitude_cyclone_abi()
        except ImportError:
            if len(get_filenames(self.subdir)) != 16:
                raise RuntimeError("Existing data files do not match the expected number of files.")
        download_rsr()
        download_luts(aerosol_types=["rayleigh_only"])

    def setup(self):
        """Set up the benchmarks."""
        import dask.config
        self.data_files = get_filenames(self.subdir)
        dask.config.set({"array.chunk-size": "32MiB"})

    def time_load_one_channel(self):
        """Time the loading of one channel."""
        self.compute_channel("C01")

    def peakmem_load_one_channel(self):
        """Check peak memory usage of loading one channel."""
        self.compute_channel("C01")

    def time_load_true_color(self):
        """Time the loading of the generation of true_color."""
        self.compute_composite("true_color")

    def peakmem_load_true_color(self):
        """Check peak memory usage of the generation of true_color."""
        self.compute_composite("true_color")

    def time_save_true_color_nocorr_to_geotiff(self):
        """Time the generation and saving of true_color_nocorr."""
        self.save_composite_as_geotiff("true_color_nocorr")

    def peakmem_save_true_color_to_geotiff(self):
        """Check peak memory usage of the generation and saving of true_color_nocorr."""
        self.save_composite_as_geotiff("true_color_nocorr")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

r"""IASI L2 SO2 BUFR format reader.

Introduction
------------
The ``iasi_l2_so2_bufr`` reader reads IASI level2 SO2 data in BUFR format. The algorithm is described in the
Theoretical Basis Document, linked below.

Each BUFR file consists of a number of messages, one for each scan, each of which contains SO2 column amounts
in Dobson units for retrievals performed with plume heights of 7, 10, 13, 16 and 25 km.

Reader Arguments
----------------
A list of retrieval files, fnames, can be opened as follows::

  Scene(reader="iasi_l2_so2_bufr", filenames=fnames)

Example:
--------
Here is an example how to read the data in satpy:

.. code-block:: python

    from satpy import Scene
    import glob

    filenames = glob.glob(
        '/test_data/W_XX-EUMETSAT-Darmstadt,SOUNDING+SATELLITE,METOPA+IASI_C_EUMC_20200204091455_68984_eps_o_so2_l2.bin')
    scn = Scene(filenames=filenames, reader='iasi_l2_so2_bufr')
    scn.load(['so2_height_3', 'so2_height_4'])
    print(scn['so2_height_3'])


Output:

.. code-block:: none

    <xarray.DataArray 'so2_height_3' (y: 23, x: 120)>
    dask.array<where, shape=(23, 120), dtype=float64, chunksize=(1, 120), chunktype=numpy.ndarray>
    Coordinates:
        crs      object +proj=latlong +datum=WGS84 +ellps=WGS84 +type=crs
    Dimensions without coordinates: y, x
    Attributes:
        sensor:               IASI
        units:                dobson
        file_type:            iasi_l2_so2_bufr
        wavelength:           None
        modifiers:            ()
        platform_name:        METOP-2
        resolution:           12000
        fill_value:           -1e+100
        level:                None
        polarization:         None
        coordinates:          ('longitude', 'latitude')
        calibration:          None
        key:                  #3#sulphurDioxide
        name:                 so2_height_3
        start_time:           2020-02-04 09:14:55
        end_time:             2020-02-04 09:17:51
        area:                 Shape: (23, 120)\nLons: <xarray.DataArray 'longitud...
        ancillary_variables:  []

References:
Algorithm Theoretical Basis Document:
https://acsaf.org/docs/atbd/Algorithm_Theoretical_Basis_Document_IASI_SO2_Jul_2016.pdf
"""


# TDB: this reader is based on iasi_l2.py and seviri_l2_bufr.py

import datetime as dt
import logging

import dask.array as da
import numpy as np
import xarray as xr

try:
    import eccodes as ec
except ImportError as e:
    raise ImportError(
        """Missing eccodes-python and/or eccodes C-library installation. Use conda to install eccodes.
           Error: """, e)

from satpy.readers.core.file_handlers import BaseFileHandler
from satpy.utils import get_legacy_chunk_size

logger = logging.getLogger("IASIL2SO2BUFR")
CHUNK_SIZE = get_legacy_chunk_size()
data_center_dict = {3: "METOP-1", 4: "METOP-2", 5: "METOP-3"}


class IASIL2SO2BUFR(BaseFileHandler):
    """File handler for the IASI L2 SO2 BUFR product."""

    def __init__(self, filename, filename_info, filetype_info, **kwargs):
        """Initialise the file handler for the IASI L2 SO2 BUFR data."""
        super(IASIL2SO2BUFR, self).__init__(filename, filename_info, filetype_info)

        start_time, end_time = self.get_start_end_date()

        sc_id = self.get_attribute("satelliteIdentifier")

        self.metadata = {}
        self.metadata["start_time"] = start_time
        self.metadata["end_time"] = end_time
        self.metadata["SpacecraftName"] = data_center_dict[sc_id]

    @property
    def start_time(self):
        """Return the start time of data acqusition."""
        return self.metadata["start_time"]

    @property
    def end_time(self):
        """Return the end time of data acquisition."""
        return self.metadata["end_time"]

    @property
    def platform_name(self):
        """Return spacecraft name."""
        return "{}".format(self.metadata["SpacecraftName"])

    def get_start_end_date(self):
        """Get the first and last date from the bufr file."""
        fh = open(self.filename, "rb")
        i = 0
        while True:
            # get handle for message
            bufr = ec.codes_bufr_new_from_file(fh)
            if bufr is None:
                break
            ec.codes_set(bufr, "unpack", 1)
            year = ec.codes_get(bufr, "year")
            month = ec.codes_get(bufr, "month")
            day = ec.codes_get(bufr, "day")
            hour = ec.codes_get(bufr, "hour")
            minute = ec.codes_get(bufr, "minute")
            second = ec.codes_get(bufr, "second")

            obs_time = dt.datetime(year=year, month=month, day=day, hour=hour, minute=minute, second=second)

            if i == 0:
                start_time = obs_time

            ec.codes_release(bufr)

            i += 1

        end_time = obs_time

        fh.close()

        return start_time, end_time

    def get_attribute(self, key):
        """Get BUFR attributes."""
        # This function is inefficient as it is looping through the entire
        # file to get 1 attribute. It causes a problem though if you break
        # from the file early - dont know why but investigating - fix later

        fh = open(self.filename, "rb")
        while True:
            # get handle for message
            bufr = ec.codes_bufr_new_from_file(fh)
            if bufr is None:
                break
            ec.codes_set(bufr, "unpack", 1)
            attr = ec.codes_get(bufr, key)
            ec.codes_release(bufr)

        fh.close()
        return attr

    def get_array(self, key):
        """Get all data from file for the given BUFR key."""
        with open(self.filename, "rb") as fh:
            msgCount = 0
            while True:

                bufr = ec.codes_bufr_new_from_file(fh)
                if bufr is None:
                    break

                ec.codes_set(bufr, "unpack", 1)

                values = ec.codes_get_array(
                        bufr, key, float)

                if len(values) == 1:
                    values = np.repeat(values, 120)

                # if is the first message initialise our final array
                if (msgCount == 0):

                    arr = da.from_array([values], chunks=CHUNK_SIZE)
                else:
                    tmpArr = da.from_array([values], chunks=CHUNK_SIZE)

                    arr = da.concatenate((arr, tmpArr), axis=0)

                msgCount = msgCount+1
                ec.codes_release(bufr)

        if arr.size == 1:
            arr = arr[0]

        return arr

    def get_dataset(self, dataset_id, dataset_info):
        """Get dataset using the BUFR key in dataset_info."""
        arr = self.get_array(dataset_info["key"])
        arr[arr == dataset_info["fill_value"]] = np.nan

        xarr = xr.DataArray(arr, dims=["y", "x"], name=dataset_info["name"])
        xarr.attrs["sensor"] = "IASI"
        xarr.attrs["platform_name"] = self.platform_name
        xarr.attrs.update(dataset_info)

        return xarr



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Tests for the VIIRS EDR Flood reader."""
import os
import unittest
from unittest import mock

import numpy as np

from satpy.tests.reader_tests.test_hdf4_utils import FakeHDF4FileHandler

DEFAULT_FILE_DTYPE = np.uint16
DEFAULT_FILE_SHAPE = (10, 300)
DEFAULT_FILE_DATA = np.arange(DEFAULT_FILE_SHAPE[0] * DEFAULT_FILE_SHAPE[1],
                              dtype=DEFAULT_FILE_DTYPE).reshape(DEFAULT_FILE_SHAPE)


class FakeHDF4FileHandler2(FakeHDF4FileHandler):
    """Swap in HDF4 file handler."""

    def get_test_content(self, filename, filename_info, filename_type):
        """Mimic reader input file content."""
        file_content = {}
        file_content["/attr/Satellitename"] = filename_info["platform_shortname"]
        file_content["/attr/SensorIdentifyCode"] = "VIIRS"

        # only one dataset for the flood reader
        file_content["WaterDetection"] = DEFAULT_FILE_DATA
        file_content["WaterDetection/attr/_Fillvalue"] = 1
        file_content["WaterDetection/attr/scale_factor"] = 1.
        file_content["WaterDetection/attr/add_offset"] = 0.
        file_content["WaterDetection/attr/units"] = "none"
        file_content["WaterDetection/shape"] = DEFAULT_FILE_SHAPE
        file_content["WaterDetection/attr/ProjectionMinLatitude"] = 15.
        file_content["WaterDetection/attr/ProjectionMaxLatitude"] = 68.
        file_content["WaterDetection/attr/ProjectionMinLongitude"] = -124.
        file_content["WaterDetection/attr/ProjectionMaxLongitude"] = -61.

        # convert tp xarrays
        from xarray import DataArray
        for key, val in file_content.items():
            if isinstance(val, np.ndarray):
                attrs = {}
                for a in ["_Fillvalue", "units", "ProjectionMinLatitude", "ProjectionMaxLongitude",
                          "ProjectionMinLongitude", "ProjectionMaxLatitude"]:
                    if key + "/attr/" + a in file_content:
                        attrs[a] = file_content[key + "/attr/" + a]
                if val.ndim > 1:
                    file_content[key] = DataArray(val, dims=("fakeDim0", "fakeDim1"), attrs=attrs)
                else:
                    file_content[key] = DataArray(val, attrs=attrs)

        if "y" not in file_content["WaterDetection"].dims:
            file_content["WaterDetection"] = file_content["WaterDetection"].rename({"fakeDim0": "x", "fakeDim1": "y"})
        return file_content


class TestVIIRSEDRFloodReader(unittest.TestCase):
    """Test VIIRS EDR Flood Reader."""

    yaml_file = "viirs_edr_flood.yaml"

    def setUp(self):
        """Wrap HDF4 file handler with own fake file handler."""
        from satpy._config import config_search_paths
        from satpy.readers.viirs_edr_flood import VIIRSEDRFlood
        self.reader_configs = config_search_paths(os.path.join("readers", self.yaml_file))
        self.p = mock.patch.object(VIIRSEDRFlood, "__bases__", (FakeHDF4FileHandler2,))
        self.fake_handler = self.p.start()
        self.p.is_local = True

    def tearDown(self):
        """Stop wrapping the HDF4 file handler."""
        self.p.stop()

    def test_init(self):
        """Test basic init with no extra parameters."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "WATER_VIIRS_Prj_SVI_npp_d20180824_t1828213_e1839433_b35361_cspp_dev_10_300_01.hdf"
        ])
        assert len(loadables) == 1
        r.create_filehandlers(loadables)
        assert r.file_handlers

    def test_load_dataset(self):
        """Test loading all datasets from a full swath file."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "WATER_VIIRS_Prj_SVI_npp_d20180824_t1828213_e1839433_b35361_cspp_dev_10_300_01.hdf"
        ])
        r.create_filehandlers(loadables)
        datasets = r.load(["WaterDetection"])
        assert len(datasets) == 1
        for v in datasets.values():
            assert v.attrs["units"] == "none"

    def test_load_dataset_aoi(self):
        """Test loading all datasets from an area of interest file."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "WATER_VIIRS_Prj_SVI_npp_d20180824_t1828213_e1839433_b35361_cspp_dev_001_10_300_01.hdf"
        ])
        r.create_filehandlers(loadables)
        datasets = r.load(["WaterDetection"])
        assert len(datasets) == 1
        for v in datasets.values():
            assert v.attrs["units"] == "none"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2018 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Module for testing the satpy.readers.clavrx module."""

import os
import unittest
from unittest import mock

import dask.array as da
import numpy as np
import xarray as xr
from pyresample.geometry import SwathDefinition

from satpy.tests.reader_tests.test_hdf4_utils import FakeHDF4FileHandler

DEFAULT_FILE_DTYPE = np.uint16
DEFAULT_FILE_SHAPE = (10, 300)
DEFAULT_FILE_DATA = np.arange(DEFAULT_FILE_SHAPE[0] * DEFAULT_FILE_SHAPE[1],
                              dtype=DEFAULT_FILE_DTYPE).reshape(DEFAULT_FILE_SHAPE)
DEFAULT_FILE_FACTORS = np.array([2.0, 1.0], dtype=np.float32)
DEFAULT_LAT_DATA = np.linspace(45, 65, DEFAULT_FILE_SHAPE[1]).astype(DEFAULT_FILE_DTYPE)
DEFAULT_LAT_DATA = np.repeat([DEFAULT_LAT_DATA], DEFAULT_FILE_SHAPE[0], axis=0)
DEFAULT_LON_DATA = np.linspace(5, 45, DEFAULT_FILE_SHAPE[1]).astype(DEFAULT_FILE_DTYPE)
DEFAULT_LON_DATA = np.repeat([DEFAULT_LON_DATA], DEFAULT_FILE_SHAPE[0], axis=0)


class FakeHDF4FileHandlerPolar(FakeHDF4FileHandler):
    """Swap-in HDF4 File Handler."""

    def get_test_content(self, filename, filename_info, filetype_info):
        """Mimic reader input file content."""
        file_content = {
            "/attr/platform": "SNPP",
            "/attr/sensor": "VIIRS",
        }
        file_content["longitude"] = xr.DataArray(
            da.from_array(DEFAULT_LON_DATA, chunks=4096),
            attrs={
                "_FillValue": np.nan,
                "scale_factor": 1.,
                "add_offset": 0.,
                "standard_name": "longitude",
            })
        file_content["longitude/shape"] = DEFAULT_FILE_SHAPE

        file_content["latitude"] = xr.DataArray(
            da.from_array(DEFAULT_LAT_DATA, chunks=4096),
            attrs={
                "_FillValue": np.nan,
                "scale_factor": 1.,
                "add_offset": 0.,
                "standard_name": "latitude",
            })
        file_content["latitude/shape"] = DEFAULT_FILE_SHAPE

        file_content["variable1"] = xr.DataArray(
            da.from_array(DEFAULT_FILE_DATA, chunks=4096).astype(np.float32),
            attrs={
                "_FillValue": -1,
                "scale_factor": 1.,
                "add_offset": 0.,
                "units": "1",
            })
        file_content["variable1/shape"] = DEFAULT_FILE_SHAPE

        # data with fill values
        file_content["variable2"] = xr.DataArray(
            da.from_array(DEFAULT_FILE_DATA, chunks=4096).astype(np.float32),
            attrs={
                "_FillValue": -1,
                "scale_factor": 1.,
                "add_offset": 0.,
                "units": "1",
            })
        file_content["variable2/shape"] = DEFAULT_FILE_SHAPE
        file_content["variable2"] = file_content["variable2"].where(
                                        file_content["variable2"] % 2 != 0)

        # category
        file_content["variable3"] = xr.DataArray(
            da.from_array(DEFAULT_FILE_DATA, chunks=4096).astype(np.byte),
            attrs={
                "SCALED": 0,
                "_FillValue": -128,
                "flag_meanings": "clear water supercooled mixed ice unknown",
                "flag_values": [0, 1, 2, 3, 4, 5],
                "units": "none",
            })
        file_content["variable3/shape"] = DEFAULT_FILE_SHAPE

        file_content["refl_1_38um_nom"] = xr.DataArray(
            da.from_array(DEFAULT_FILE_DATA, chunks=4096).astype(np.float32),
            attrs={
                "SCALED": 1,
                "add_offset": 59.0,
                "scale_factor": 0.0018616290763020515,
                "units": "%",
                "_FillValue": -32768,
                "valid_range": [-32767, 32767],
                "actual_range": [-2., 120.],
                "actual_missing": -999.0
            })
        file_content["refl_1_38um_nom/shape"] = DEFAULT_FILE_SHAPE

        return file_content


class TestCLAVRXReaderPolar(unittest.TestCase):
    """Test CLAVR-X Reader with Polar files."""

    yaml_file = "clavrx.yaml"

    def setUp(self):
        """Wrap HDF4 file handler with our own fake handler."""
        from satpy._config import config_search_paths
        from satpy.readers.clavrx import CLAVRXHDF4FileHandler
        self.reader_configs = config_search_paths(os.path.join("readers", self.yaml_file))
        # http://stackoverflow.com/questions/12219967/how-to-mock-a-base-class-with-python-mock-library
        self.p = mock.patch.object(CLAVRXHDF4FileHandler, "__bases__", (FakeHDF4FileHandlerPolar,))
        self.fake_handler = self.p.start()
        self.p.is_local = True

    def tearDown(self):
        """Stop wrapping the NetCDF4 file handler."""
        self.p.stop()

    def test_init(self):
        """Test basic init with no extra parameters."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "clavrx_npp_d20170520_t2053581_e2055223_b28822.level2.hdf",
        ])
        assert len(loadables) == 1
        r.create_filehandlers(loadables)
        # make sure we have some files
        assert r.file_handlers

    def test_available_datasets(self):
        """Test available_datasets with fake variables from YAML."""
        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        loadables = r.select_files_from_pathnames([
            "clavrx_npp_d20170520_t2053581_e2055223_b28822.level2.hdf",
        ])
        assert len(loadables) == 1
        r.create_filehandlers(loadables)
        # make sure we have some files
        assert r.file_handlers

        # mimic the YAML file being configured for more datasets
        fake_dataset_info = [
            (None, {"name": "variable1", "resolution": None, "file_type": ["clavrx_hdf4"]}),
            (True, {"name": "variable2", "resolution": 742, "file_type": ["clavrx_hdf4"]}),
            (True, {"name": "variable2", "resolution": 1, "file_type": ["clavrx_hdf4"]}),
            (None, {"name": "variable2", "resolution": 1, "file_type": ["clavrx_hdf4"]}),
            (None, {"name": "_fake1", "file_type": ["clavrx_hdf4"]}),
            (None, {"name": "variable1", "file_type": ["level_fake"]}),
            (True, {"name": "variable3", "file_type": ["clavrx_hdf4"]}),
        ]
        new_ds_infos = list(r.file_handlers["clavrx_hdf4"][0].available_datasets(
            fake_dataset_info))
        assert len(new_ds_infos) == 9

        # we have this and can provide the resolution
        assert new_ds_infos[0][0]
        assert new_ds_infos[0][1]["resolution"] == 742  # hardcoded

        # we have this, but previous file handler said it knew about it
        # and it is producing the same resolution as what we have
        assert new_ds_infos[1][0]
        assert new_ds_infos[1][1]["resolution"] == 742

        # we have this, but don't want to change the resolution
        # because a previous handler said it has it
        assert new_ds_infos[2][0]
        assert new_ds_infos[2][1]["resolution"] == 1

        # even though the previous one was known we can still
        # produce it at our new resolution
        assert new_ds_infos[3][0]
        assert new_ds_infos[3][1]["resolution"] == 742

        # we have this and can update the resolution since
        # no one else has claimed it
        assert new_ds_infos[4][0]
        assert new_ds_infos[4][1]["resolution"] == 742

        # we don"t have this variable, don't change it
        assert not new_ds_infos[5][0]
        assert new_ds_infos[5][1].get("resolution") is None

        # we have this, but it isn't supposed to come from our file type
        assert new_ds_infos[6][0] is None
        assert new_ds_infos[6][1].get("resolution") is None

        # we could have loaded this but some other file handler said it has this
        assert new_ds_infos[7][0]
        assert new_ds_infos[7][1].get("resolution") is None

        # we can add resolution to the previous dataset, so we do
        assert new_ds_infos[8][0]
        assert new_ds_infos[8][1]["resolution"] == 742

    def test_available_datasets_with_alias(self):
        """Test availability of aliased dataset."""
        import xarray as xr

        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        with mock.patch("satpy.readers.clavrx.SDS", xr.DataArray):
            loadables = r.select_files_from_pathnames([
                "clavrx_npp_d20170520_t2053581_e2055223_b28822.level2.hdf",
            ])
            r.create_filehandlers(loadables)
            available_ds = list(r.file_handlers["clavrx_hdf4"][0].available_datasets())

            assert available_ds[5][1]["name"] == "refl_1_38um_nom"
            assert available_ds[6][1]["name"] == "M09"
            assert available_ds[6][1]["file_key"] == "refl_1_38um_nom"

    def test_load_all(self):
        """Test loading all test datasets."""
        import xarray as xr

        from satpy.readers.core.loading import load_reader
        r = load_reader(self.reader_configs)
        with mock.patch("satpy.readers.clavrx.SDS", xr.DataArray):
            loadables = r.select_files_from_pathnames([
                "clavrx_npp_d20170520_t2053581_e2055223_b28822.level2.hdf",
            ])
            r.create_filehandlers(loadables)

        var_list = ["M09", "variable2", "variable3"]
        datasets = r.load(var_list)
        assert len(datasets) == len(var_list)
        for v in datasets.values():
            assert v.attrs["units"] in ["1", "%"]
            assert v.attrs["platform_name"] == "npp"
            assert v.attrs["sensor"] == "viirs"
            assert isinstance(v.attrs["area"], SwathDefinition)
            assert v.attrs["area"].lons.attrs["rows_per_scan"] == 16
            assert v.attrs["area"].lats.attrs["rows_per_scan"] == 16
        assert isinstance(datasets["variable3"].attrs.get("flag_meanings"), list)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2019 Satpy developers
#
# satpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# satpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with satpy.  If not, see <http://www.gnu.org/licenses/>.

"""EUM L2 GRIB-reader test package."""

import datetime
import sys
from unittest import mock

import numpy as np
import pytest

from satpy.tests.utils import make_dataid

# Dictionary to be used as fake GRIB message
FAKE_SEVIRI_MESSAGE = {
    "longitudeOfSubSatellitePointInDegrees": 9.5,
    "dataDate": 20191020,
    "dataTime": 1745,
    "Nx": 1000,
    "Ny": 1200,
    "earthMajorAxis": 6400.,
    "earthMinorAxis": 6300.,
    "NrInRadiusOfEarth": 6.,
    "XpInGridLengths": 500,
    "parameterNumber": 30,
    "missingValue": 9999,
}

FAKE_FCI_MESSAGE = {
    "longitudeOfSubSatellitePointInDegrees": 0.0,
    "dataDate": 20191020,
    "dataTime": 1745,
    "Nx": 5568,
    "Ny": 5568,
    "earthMajorAxis": 6378140.,
    "earthMinorAxis": 6356755.,
    "NrInRadiusOfEarth": 6.6107,
    "XpInGridLengths": 2784.0,
    "parameterNumber": 30,
    "missingValue": 9999,
}

# List to be used as fake GID source
FAKE_GID = [0, 1, 2, 3, None]


@pytest.fixture
@mock.patch("satpy.readers.eum_l2_grib.ec")
def setup_reader(ec_):
    """Set up the test by creating a mocked eccodes library."""
    fake_gid_generator = (i for i in FAKE_GID)
    ec_.codes_grib_new_from_file.side_effect = lambda fh: next(fake_gid_generator)
    return ec_


def common_checks(ec_, reader, mock_file, dataset_id):
    """Commmon checks for fci and seviri data."""
    # Checks that the codes_grib_multi_support_on function has been called
    ec_.codes_grib_multi_support_on.assert_called()

    # Restarts the id generator and clears the call history
    fake_gid_generator = (i for i in FAKE_GID)
    ec_.codes_grib_new_from_file.side_effect = lambda fh: next(fake_gid_generator)
    ec_.codes_grib_new_from_file.reset_mock()
    ec_.codes_release.reset_mock()

    # Checks the correct execution of the get_dataset function with a valid parameter_number
    valid_dataset = reader.get_dataset(dataset_id, {"parameter_number": 30})
    # Checks the correct file open call
    mock_file.assert_called_with("test.grib", "rb")
    # Checks that the dataset has been created as a DataArray object
    assert valid_dataset._extract_mock_name() == "xr.DataArray()"
    # Checks that codes_release has been called after each codes_grib_new_from_file call
    # (except after the last one which has returned a None)
    assert ec_.codes_grib_new_from_file.call_count == ec_.codes_release.call_count + 1

    # Restarts the id generator and clears the call history
    fake_gid_generator = (i for i in FAKE_GID)
    ec_.codes_grib_new_from_file.side_effect = lambda fh: next(fake_gid_generator)
    ec_.codes_grib_new_from_file.reset_mock()
    ec_.codes_release.reset_mock()

    # Checks the correct execution of the get_dataset function with an invalid parameter_number
    invalid_dataset = reader.get_dataset(dataset_id, {"parameter_number": 50})
    # Checks that the function returns None
    assert invalid_dataset is None
    # Checks that codes_release has been called after each codes_grib_new_from_file call
    # (except after the last one which has returned a None)
    assert ec_.codes_grib_new_from_file.call_count == ec_.codes_release.call_count + 1


@pytest.mark.skipif(sys.platform.startswith("win"), reason="'eccodes' not supported on Windows")
@mock.patch("satpy.readers.eum_l2_grib.xr")
@mock.patch("satpy.readers.eum_l2_grib.da")
def test_seviri_data_reading(da_, xr_, setup_reader):
    """Test the reading of data from the product."""
    from satpy.readers.eum_l2_grib import EUML2GribFileHandler
    from satpy.utils import get_legacy_chunk_size
    ec_ = setup_reader
    chunk_size = get_legacy_chunk_size()

    with mock.patch("builtins.open", mock.mock_open()) as mock_file:
        with mock.patch("satpy.readers.eum_l2_grib.ec", ec_):
            ec_.codes_get_values.return_value = np.ones(1000 * 1200)
            ec_.codes_get.side_effect = lambda gid, key: FAKE_SEVIRI_MESSAGE[key]
            reader = EUML2GribFileHandler(
                filename="test.grib",
                filename_info={
                    "spacecraft": "MET11",
                    "start_time": datetime.datetime(year=2020, month=10, day=20,
                                                    hour=19, minute=45, second=0)
                },
                filetype_info={
                    "file_type": "seviri"
                }
            )

            dataset_id = make_dataid(name="dummmy", resolution=3000)

            # Check that end_time is None for SEVIRI before the dataset has been loaded
            assert reader.end_time is None

            common_checks(ec_, reader, mock_file, dataset_id)

            # Check that end_time is now a valid datetime.datetime object after the dataset has been loaded
            assert reader.end_time == datetime.datetime(year=2020, month=10, day=20,
                                                        hour=19, minute=50, second=0)


            # Checks the correct execution of the _get_global_attributes and _get_metadata_from_msg functions
            attributes = reader._get_attributes()
            expected_attributes = {
                "orbital_parameters": {
                    "projection_longitude": 9.5
                },
                "sensor": "seviri",
                "platform_name": "Meteosat-11"
            }
            assert attributes == expected_attributes

            # Checks the reading of an array from the message
            reader._get_xarray_from_msg(0)

            # Checks that dask.array has been called with the correct arguments
            name, args, kwargs = da_.mock_calls[0]
            assert np.all(args[0] == np.ones((1200, 1000)))
            assert args[1] == chunk_size

            # Checks that xarray.DataArray has been called with the correct arguments
            name, args, kwargs = xr_.mock_calls[0]
            assert kwargs["dims"] == ("y", "x")

            # Checks the correct execution of the _get_proj_area function
            pdict, area_dict = reader._get_proj_area(0)

            expected_pdict = {
                "a": 6400000.,
                "b": 6300000.,
                "h": 32000000.,
                "ssp_lon": 9.5,
                "nlines": 1000,
                "ncols": 1200,
                "a_name": "msg_seviri_rss_3km",
                "a_desc": "MSG SEVIRI Rapid Scanning Service area definition with 3 km resolution",
                "p_id": "",
            }
            assert pdict == expected_pdict
            expected_area_dict = {
                "center_point": 500,
                "north": 1200,
                "east": 1,
                "west": 1000,
                "south": 1,
            }
            assert area_dict == expected_area_dict

            # Checks the correct execution of the get_area_def function
            with mock.patch("satpy.readers.eum_l2_grib.seviri_calculate_area_extent",
                            mock.Mock(name="seviri_calculate_area_extent")) as cae:
                with mock.patch("satpy.readers.eum_l2_grib.get_area_definition", mock.Mock()) as gad:
                    dataset_id = make_dataid(name="dummmy", resolution=400.)
                    reader.get_area_def(dataset_id)
                    # Asserts that seviri_calculate_area_extent has been called with the correct arguments
                    expected_args = ({"center_point": 500, "east": 1, "west": 1000, "south": 1, "north": 1200,
                                      "column_step": 400., "line_step": 400.},)
                    name, args, kwargs = cae.mock_calls[0]
                    assert args == expected_args
                    # Asserts that get_area_definition has been called with the correct arguments
                    name, args, kwargs = gad.mock_calls[0]
                    assert args[0] == expected_pdict
                    # The second argument must be the return result of seviri_calculate_area_extent
                    assert args[1]._extract_mock_name() == "seviri_calculate_area_extent()"


@pytest.mark.skipif(sys.platform.startswith("win"), reason="'eccodes' not supported on Windows")
@mock.patch("satpy.readers.eum_l2_grib.xr")
@mock.patch("satpy.readers.eum_l2_grib.da")
def test_fci_data_reading(da_, xr_, setup_reader):
    """Test the reading of fci data from the product."""
    from satpy.readers.eum_l2_grib import EUML2GribFileHandler
    from satpy.utils import get_legacy_chunk_size
    ec_ = setup_reader
    chunk_size = get_legacy_chunk_size()

    with mock.patch("builtins.open", mock.mock_open()) as mock_file:
        with mock.patch("satpy.readers.eum_l2_grib.ec", ec_):
            ec_.codes_get_values.return_value = np.ones(5568 * 5568)
            ec_.codes_get.side_effect = lambda gid, key: FAKE_FCI_MESSAGE[key]
            reader = EUML2GribFileHandler(
                filename="test.grib",
                filename_info={
                    "spacecraft_id": "1",
                    "start_time": datetime.datetime(year=2020, month=10, day=20,
                                                    hour=19, minute=40, second=0),
                    "end_time": datetime.datetime(year=2020, month=10, day=20,
                                                  hour=19, minute=50, second=0)
                },
                filetype_info={
                    "file_type": "fci"
                }
            )

            dataset_id = make_dataid(name="dummmy", resolution=2000)

            # Check end_time
            assert reader.end_time == datetime.datetime(year=2020, month=10, day=20,
                                                        hour=19, minute=50, second=0)

            common_checks(ec_, reader, mock_file, dataset_id)

            # Checks the correct execution of the _get_global_attributes and _get_metadata_from_msg functions
            attributes = reader._get_attributes()
            expected_attributes = {
                "orbital_parameters": {
                    "projection_longitude": 0.0
                },
                "sensor": "fci",
                "platform_name": "MTG-i1"
            }
            assert attributes == expected_attributes

            # Checks the reading of an array from the message
            reader._get_xarray_from_msg(0)

            # Checks that dask.array has been called with the correct arguments
            name, args, kwargs = da_.mock_calls[0]
            assert np.all(args[0] == np.ones((5568, 5568)))
            assert args[1] == chunk_size

            # Checks that xarray.DataArray has been called with the correct arguments
            name, args, kwargs = xr_.mock_calls[0]
            assert kwargs["dims"] == ("y", "x")

            # Checks the correct execution of the _get_proj_area function
            pdict, area_dict = reader._get_proj_area(0)

            expected_pdict = {
                "a": 6378140000.0,
                "b": 6356755000.0,
                "h": 35785830098.0,
                "ssp_lon": 0.0,
                "nlines": 5568,
                "ncols": 5568,
                "a_name": "msg_fci_fdss_2km",
                "a_desc": "MSG FCI Full Disk Scanning Service area definition with 2 km resolution",
                "p_id": ""
            }
            assert pdict == expected_pdict
            expected_area_dict = {
                "nlines": 5568,
                "ncols": 5568
            }
            assert area_dict == expected_area_dict

            # Checks the correct execution of the get_area_def function
            with mock.patch("satpy.readers.eum_l2_grib.fci_calculate_area_extent",
                            mock.Mock(name="fci_calculate_area_extent")) as cae:
                with mock.patch("satpy.readers.eum_l2_grib.get_area_definition", mock.Mock()) as gad:
                    dataset_id = make_dataid(name="dummmy", resolution=2000.)
                    reader.get_area_def(dataset_id)
                    # Asserts that seviri_calculate_area_extent has been called with the correct arguments
                    expected_args = ({"nlines": 5568, "ncols": 5568,
                                      "column_step": 2000., "line_step": 2000.},)
                    name, args, kwargs = cae.mock_calls[0]
                    assert args == expected_args
                    # Asserts that get_area_definition has been called with the correct arguments
                    name, args, kwargs = gad.mock_calls[0]
                    assert args[0] == expected_pdict
                    # The second argument must be the return result of seviri_calculate_area_extent
                    assert args[1]._extract_mock_name() == "fci_calculate_area_extent()"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2018 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Module for testing the satpy.readers.nc_slstr module."""

import datetime as dt
import unittest
import unittest.mock as mock

import numpy as np
import pytest
import xarray as xr

from satpy.dataset.dataid import DataID, ModifierTuple, WavelengthRange
from satpy.readers.slstr_l1b import NCSLSTR1B, NCSLSTRAngles, NCSLSTRFlag, NCSLSTRGeo

local_id_keys_config = {"name": {
    "required": True,
},
    "wavelength": {
        "type": WavelengthRange,
    },
    "resolution": None,
    "calibration": {
        "enum": [
            "reflectance",
            "brightness_temperature",
            "radiance",
            "counts"
        ]
    },
    "stripe": {
        "enum": [
            "a",
            "b",
            "c",
            "i",
            "f",
        ]
    },
    "view": {
        "enum": [
            "nadir",
            "oblique",
        ]
    },
    "modifiers": {
        "required": True,
        "default": ModifierTuple(),
        "type": ModifierTuple,
    },
}


class TestSLSTRL1B(unittest.TestCase):
    """Common setup for SLSTR_L1B tests."""

    @mock.patch("satpy.readers.slstr_l1b.xr")
    def setUp(self, xr_):
        """Create a fake dataset using the given radiance data."""
        self.base_data = np.array(([1., 2., 3., 4., 5., 6., 7., 8., 9.],
                                   [7., 8., 9., 10., 11., 12., 13., 14., 15.],
                                   [16., 17., 18., 19., 20., 21., 22., 23., 24.],
                                   [25., 26., 27., 28., 29., 30., 31., 32., 33.]))
        self.ang_data = np.array(([345., 355., 1., 5., 10.],
                                  [346., 355., 0.5, 4.2, 15.],
                                  [342., 356., 1., 5.3, 12.],
                                  [344.3, 356.1, 0.0001, 4.9, 9.2],))

        self.tx_data = np.array(([101000, 100500, 100000, 99500, 99000],
                                 [101000, 100500, 100000, 99500, 99000],
                                 [101000, 100500, 100000, 99500, 99000],
                                 [101000, 100500, 100000, 99500, 99000]))
        self.ty_data = np.array(([199500, 199500, 199500, 199500, 199500],
                                 [200000, 200000, 200000, 200000, 200000],
                                 [200500, 200500, 200500, 200500, 200500],
                                 [201000, 201000, 201000, 201000, 201000],))

        self.ix_data = np.array(([100800, 100600, 100400, 100200, 100000, 99800, 99600, 99400, 99200],
                                 [100800, 100600, 100400, 100200, 100000, 99800, 99600, 99400, 99200],
                                 [100800, 100600, 100400, 100200, 100000, 99800, 99600, 99400, 99200],
                                 [100800, 100600, 100400, 100200, 100000, 99800, 99600, 99400, 99200]))
        self.iy_data = np.array(([199800, 199800, 199800, 199800, 199800, 199800, 199800, 199800, 199800],
                                 [200000, 200000, 200000, 200000, 200000, 200000, 200000, 200000, 200000],
                                 [200600, 200600, 200600, 200600, 200600, 200600, 200600, 200600, 200600],
                                 [200800, 200800, 200800, 200800, 200800, 200800, 200800, 200800, 200800]))

        self.det_data = np.array(([0, 1, 1, 0, 1, 0, 1, 0, 1],
                                  [1, 0, 0, 0, 0, 1, 0, 1, 1],
                                  [1, 1, 1, 1, 1, 1, 1, 0, 1],
                                  [0, 1, 1, 0, 1, 1, 1, 1, 0]))
        self.start_time = "2020-05-10T12:01:15.585Z"
        self.end_time = "2020-05-10T12:06:18.012Z"

        self.rad = xr.DataArray(
            self.base_data,
            dims=("columns", "rows"),
            attrs={"scale_factor": 1.0, "add_offset": 0.0,
                   "_FillValue": -32768, "units": "mW.m-2.sr-1.nm-1",})
        det = xr.DataArray(
            self.base_data,
            dims=("columns", "rows"),
            attrs={"scale_factor": 1.0, "add_offset": 0.0,"_FillValue": 255,})
        x_in = xr.DataArray(self.ix_data,dims=("columns", "rows"))
        y_in = xr.DataArray(self.iy_data,dims=("columns", "rows"))

        saa = xr.DataArray(self.ang_data,dims=("columns_t", "rows_t"))
        x_tx = xr.DataArray(self.tx_data,dims=("columns_t", "rows_t"))
        y_tx = xr.DataArray(self.ty_data,dims=("columns_t", "rows_t"))

        self.fake_dataset = xr.Dataset(
            data_vars={
                "S5_radiance_an": self.rad,
                "S9_BT_ao": self.rad,
                "foo_radiance_an": self.rad,
                "S5_solar_irradiances": self.rad,
                "geometry_tn": self.rad,
                "latitude_an": self.rad,
                "x_in": x_in,
                "y_in": y_in,
                "x_an": x_in,
                "y_an": y_in,
                "flags_an": self.rad,
                "detector_an": det,
                "x_tx": x_tx,
                "y_tx": y_tx,
                "solar_azimuth_tn": saa,
            },
            attrs={
                "start_time": self.start_time,
                "stop_time": self.end_time,
            },
        )


def make_dataid(**items):
    """Make a data id."""
    return DataID(local_id_keys_config, **items)


class TestSLSTRReader(TestSLSTRL1B):
    """Test various nc_slstr file handlers."""

    class FakeSpl:
        """Fake return function for SPL interpolation."""

        @staticmethod
        def ev(foo_x, foo_y):
            """Fake function to return interpolated data."""
            return np.zeros((3, 2))

    @mock.patch("satpy.readers.slstr_l1b.xr")
    @mock.patch("scipy.interpolate.RectBivariateSpline")
    def test_instantiate(self, bvs_, xr_):
        """Test initialization of file handlers."""
        bvs_.return_value = self.FakeSpl
        xr_.open_dataset.return_value = self.fake_dataset

        good_start = dt.datetime.strptime(self.start_time,
                                          "%Y-%m-%dT%H:%M:%S.%fZ")
        good_end = dt.datetime.strptime(self.end_time,
                                        "%Y-%m-%dT%H:%M:%S.%fZ")

        ds_id = make_dataid(name="foo", calibration="radiance",
                            stripe="a", view="nadir")
        ds_id_500 = make_dataid(name="foo", calibration="radiance",
                                stripe="a", view="nadir", resolution=500)
        filename_info = {"mission_id": "S3A", "dataset_name": "foo",
                         "start_time": 0, "end_time": 0,
                         "stripe": "a", "view": "n"}
        test = NCSLSTR1B("somedir/S1_radiance_an.nc", filename_info, "c")
        assert test.view == "nadir"
        assert test.stripe == "a"
        with pytest.warns(UserWarning, match=r"No radiance adjustment supplied for channel"):
            test.get_dataset(ds_id, dict(filename_info, **{"file_key": "foo"}))
        assert test.start_time == good_start
        assert test.end_time == good_end
        xr_.open_dataset.assert_called()
        xr_.open_dataset.reset_mock()

        filename_info = {"mission_id": "S3A", "dataset_name": "foo",
                         "start_time": 0, "end_time": 0,
                         "stripe": "c", "view": "o"}
        test = NCSLSTR1B("somedir/S1_radiance_co.nc", filename_info, "c")
        assert test.view == "oblique"
        assert test.stripe == "c"
        test.get_dataset(ds_id, dict(filename_info, **{"file_key": "foo"}))
        assert test.start_time == good_start
        assert test.end_time == good_end
        xr_.open_dataset.assert_called()
        xr_.open_dataset.reset_mock()

        filename_info = {"mission_id": "S3A", "dataset_name": "foo",
                         "start_time": 0, "end_time": 0,
                         "stripe": "a", "view": "n"}
        test = NCSLSTRGeo("somedir/geometry_an.nc", filename_info, "c")
        test.get_dataset(ds_id, dict(filename_info, **{"file_key": "latitude_{stripe:1s}{view:1s}"}))
        assert test.start_time == good_start
        assert test.end_time == good_end
        xr_.open_dataset.assert_called()
        xr_.open_dataset.reset_mock()

        test = NCSLSTRFlag("somedir/S1_radiance_an.nc", filename_info, "c")
        test.get_dataset(ds_id, dict(filename_info, **{"file_key": "flags_{stripe:1s}{view:1s}"}))
        assert test.view == "nadir"
        assert test.stripe == "a"
        assert test.start_time == good_start
        assert test.end_time == good_end
        xr_.open_dataset.assert_called()
        xr_.open_dataset.reset_mock()

        test = NCSLSTRAngles("somedir/S1_radiance_an.nc", filename_info, "c")
        test.get_dataset(ds_id, dict(filename_info, **{"file_key": "geometry_t{view:1s}"}))
        assert test.start_time == good_start
        assert test.end_time == good_end
        xr_.open_dataset.assert_called()
        xr_.open_dataset.reset_mock()
        test.get_dataset(ds_id_500, dict(filename_info, **{"file_key": "geometry_t{view:1s}"}))


class TestSLSTRCalibration(TestSLSTRL1B):
    """Test the implementation of the calibration factors."""

    @mock.patch("satpy.readers.slstr_l1b.xr")
    def test_radiance_calibration(self, xr_):
        """Test radiance calibration steps."""
        from satpy.readers.slstr_l1b import CHANCALIB_FACTORS
        xr_.open_dataset.return_value = self.fake_dataset

        ds_id = make_dataid(name="foo", calibration="radiance",
                            stripe="a", view="nadir")
        filename_info = {"mission_id": "S3A", "dataset_name": "foo",
                         "start_time": 0, "end_time": 0,
                         "stripe": "a", "view": "n"}

        test = NCSLSTR1B("somedir/S1_radiance_co.nc", filename_info, "c")
        # Check warning is raised if we don't have calibration
        with pytest.warns(UserWarning, match=r"No radiance adjustment supplied for channel"):
            test.get_dataset(ds_id, dict(filename_info, **{"file_key": "foo"}))

        # Check user calibration is used correctly
        test = NCSLSTR1B("somedir/S1_radiance_co.nc", filename_info, "c",
                         user_calibration={"foo_nadir": 0.4})
        data = test.get_dataset(ds_id, dict(filename_info, **{"file_key": "foo"}))
        np.testing.assert_allclose(data.values, self.base_data * 0.4)

        # Check internal calibration is used correctly
        ds_id = make_dataid(name="S5", calibration="radiance", stripe="a", view="nadir")
        filename_info["dataset_name"] = "S5"
        test = NCSLSTR1B("somedir/S1_radiance_an.nc", filename_info, "c")
        data = test.get_dataset(ds_id, dict(filename_info, **{"file_key": "S5"}))
        np.testing.assert_allclose(data.values,
                                   self.base_data * CHANCALIB_FACTORS["S5_nadir"])

    @mock.patch("satpy.readers.slstr_l1b.xr")
    @mock.patch("satpy.readers.slstr_l1b.da")
    def test_reflectance_calibration(self, da_, xr_):
        """Test reflectance calibration."""
        xr_.open_dataset.return_value = self.fake_dataset
        da_.map_blocks.return_value = self.rad / 100.
        filename_info = {"mission_id": "S3A", "dataset_name": "S5",
                         "start_time": 0, "end_time": 0,
                         "stripe": "a", "view": "n"}
        ds_id = make_dataid(name="S5", calibration="reflectance", stripe="a", view="nadir")
        test = NCSLSTR1B("somedir/S1_radiance_an.nc", filename_info, "c")
        data = test.get_dataset(ds_id, dict(filename_info, **{"file_key": "S5"}))
        assert data.units == "%"
        np.testing.assert_allclose(data.values, self.rad * np.pi)

    def test_cal_rad(self):
        """Test the radiance to reflectance converter."""
        rad = np.array([10., 20., 30., 40., 50., 60., 70.])
        didx = np.array([1, 2., 1., 3., 2., 2., 0.])
        solflux = np.array([100., 200., 300., 400.])

        good_rad = np.array([1. / 20., 1. / 15., 3. / 20., 1. / 10., 1. / 6., 2. / 10., 7. / 10.])

        out_rad = NCSLSTR1B._cal_rad(rad, didx, solflux)
        np.testing.assert_allclose(out_rad, good_rad)


class TestSLSTRAngles(TestSLSTRL1B):
    """Test the implementation of the angle reconstruction."""

    @mock.patch("satpy.readers.slstr_l1b.xr.open_dataset")
    def test_radiance_calibration(self, xr_):
        """Test radiance calibration steps."""
        xr_.return_value = self.fake_dataset

        res = np.array([[350.17659522, 353.31335956, 356.1441152, 358.55896366,
                         0.4400064, 1.79903533, 3.18424353, 5.28496166, 8.80226547],
                        [350.02717525, 353.48225908, 356.37905054, 358.72059252,
                         0.5, 1.81840943, 3.23819125, 5.44552178, 9.13958444],
                        [349.30986294, 354.38594648, 357.53203205, 359.49732861,
                         1.00001215, 2.61842911, 4.45389453, 6.49348929, 8.72979111],
                        [349.68116993, 354.62824808, 357.55405421, 359.31582013,
                         0.74004914, 2.46431559, 4.44154365, 6.45655991, 8.30055946]])

        ds_id = make_dataid(name="solar_azimuth_angle", view="nadir")
        filename_info = {"mission_id": "S3A", "dataset_name": "solar_azimuth_angle",
                         "start_time": 0, "end_time": 0,
                         "stripe": "t", "view": "n"}

        test = NCSLSTRAngles("somedir/geometry_tn.nc", filename_info, "c")
        data = test.get_dataset(ds_id, dict(filename_info, **{"file_key": "solar_azimuth_tn"}))
        assert data.attrs["units"] == "degrees"
        assert data.shape == res.shape
        np.testing.assert_allclose(data.values, res)
